/* bench-Circuit.c
 */

#include "bench-common.h"
#include <Padico/NetAccess.h>
#include <Padico/PadicoTM.h>
#include <Padico/Circuit.h>
#include <Padico/Module.h>

PADICO_MODULE_DECLARE(bench_Circuit);
PADICO_MODULE_INIT(bench_circuit_init);
PADICO_MODULE_RUN(bench_circuit_run);

static char*    recv_buffer = NULL; /**< buffer for receipt */
static size_t   recv_lenbuf = 0;    /**< len of above buffer */
static size_t   len_offset  = 0;
static marcel_t recv_tid    = NULL;
static int      num_trips   = 0;

static marcel_mutex_t mutex;
static marcel_cond_t cond;
static padico_circuit_t circuit = NULL;
static padico_group_t   group = NULL;


static void bc_send(char*buffer, size_t lenbuf)
{
  const size_t hlen = len_offset;
  const size_t blen = lenbuf - hlen;
  uint32_t header[2] = { [0] = htonl(lenbuf), [1] = htonl(num_trips) };
  padico_circuit_connection_t out_conn = padico_circuit_begin_packing(circuit, (1-padico_circuit_rank(circuit)));
  padico_circuit_pack(out_conn, header, sizeof(uint32_t)*2,
		      padico_circuit_send_CHEAPER, padico_circuit_receive_EXPRESS);
  if(blen > 0)
    {
      padico_circuit_pack(out_conn, buffer, blen, 
			  padico_circuit_send_CHEAPER, padico_circuit_receive_CHEAPER);
    }
  padico_circuit_end_packing(out_conn);
  padico_trace("send i=%d end\n", num_trips);
}


static void circuit_callback(padico_circuit_connection_t in_conn, void*key)
{
  uint32_t header[2];
  padico_out(40, "bench-Circuit: circuit_callback() received=%d\n", num_trips);
  padico_circuit_unpack(in_conn, header, sizeof(uint32_t)*2,
			padico_circuit_send_CHEAPER, padico_circuit_receive_EXPRESS);
  const uint32_t len   = ntohl(header[0]);
  const uint32_t trips = ntohl(header[1]);
  marcel_mutex_lock(&mutex);
  while((!recv_buffer) || (len != recv_lenbuf))
    {
      marcel_cond_wait(&cond, &mutex);
    }
  if(len > len_offset)
    {
      padico_circuit_unpack(in_conn, recv_buffer, len - len_offset,
			    padico_circuit_send_CHEAPER, padico_circuit_receive_CHEAPER);
    }
  padico_circuit_end_unpacking(in_conn);
  num_trips++;
  const marcel_t tid = recv_tid;
  recv_tid = NULL;
  recv_buffer = NULL;
  marcel_cond_broadcast(&cond);
  marcel_mutex_unlock(&mutex);
  marcel_yield_to(tid);
}

int bench_circuit_init(void)
{
  const char*group_name = "Group:bench-Circuit";
  const char*circuit_name = "Circuit:bench-Circuit";
  marcel_mutex_init(&mutex, NULL);
  marcel_cond_init(&cond, NULL);
  group = padico_group_init(group_name);
  padico_group_node_add(group, bench_common_param.server_node);
  padico_group_node_add(group, bench_common_param.client_node);
  padico_group_store(group);
  circuit = padico_circuit_init(group_name, circuit_name);
  padico_circuit_start(circuit, &circuit_callback, NULL);
  padico_tm_msleep(200);
  return 0;
}

static void bc_start(char*buffer, int lenbuf)
{
  marcel_mutex_lock(&mutex);
  recv_buffer = buffer;
  recv_lenbuf = lenbuf;
  recv_tid    = marcel_self();
  len_offset  = 48;
  marcel_mutex_unlock(&mutex);
  marcel_cond_signal(&cond);
}

static void bc_wait(void)
{
  padico_out(50, "entering trips=%d.\n", num_trips);
  padico_out(60, " -> locking...\n");
  marcel_mutex_lock(&mutex);
  while(recv_buffer)
    {
      padico_out(20, "waiting synchro trips=%d...\n", num_trips);
      marcel_cond_wait(&cond, &mutex);
      padico_out(20, "unlocked synchro trips=%d...\n", num_trips);
    }
  recv_lenbuf = 0;
  recv_buffer = NULL;
  marcel_mutex_unlock(&mutex);
  padico_out(60, " <- unlocked.\n");
}


static void circuit_server(char*buffer, size_t lenbuf)
{
  padico_out(20, "starting...\n");
  bc_start(buffer, lenbuf);
  bc_wait();
  bc_send(buffer, lenbuf);
}

static void circuit_client(char*buffer, size_t lenbuf)
{
  padico_out(20, "starting...\n");
  bc_start(buffer, lenbuf);
  bc_send(buffer, lenbuf);
  bc_wait();
}

static struct bench_desc_s circuit_bench =
{
  "Circuit",
  &circuit_server,
  &circuit_client
};

int bench_circuit_run(int argc, char**argv)
{
  the_bench(&circuit_bench);
  return 0;
}
