/** @file
 * @brief Circuit module external API.
 * @ingroup Circuit
 */

#ifndef PADICO_CIRCUIT_H
#define PADICO_CIRCUIT_H

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/NetSelector.h>
#include <Padico/PadicoControl.h>

/** @defgroup CircuitAPI API: Circuit -- management and messaging
 * @ingroup Circuit
 */

#ifdef __cplusplus
extern "C" {
#endif
  
#include <Padico/PM2.h>
  
  /** @addtogroup CircuitAPI
   * @{
   */
  
  /* *** CIRCUIT **************************************************** */

  /** Circuit connection (in/out)
   */
  typedef struct padico_circuit_connection_s*padico_circuit_connection_t;

  /** Circuit reception callback function
   */
  typedef void (*padico_circuit_cb_t)(padico_circuit_connection_t, void*);

  /** Circuit connection description- note: this is used for both in and out connections
   */
  struct padico_circuit_connection_s
  {
    struct padico_circuit_s*circuit; /**< circuit that owns this connection */
    puk_adapter_t assembly_pattern;
    int peer;                        /**< peer node for this connection, in group numbering */
    marcel_mutex_t lock;             /**< only one thread per connection at the same time! */
    long adapter_conn_info[8];   /**< this field contains adapter-specific data, up to 8 words */
  };
  /** Initialize a blank circuit connection
   */
  static inline void padico_circuit_connection_init(struct padico_circuit_connection_s*_c,
						    struct padico_circuit_s*_circuit,
						    puk_adapter_t _assembly,
						    int _peer)
  {
    _c->circuit          = _circuit;
    _c->assembly_pattern = _assembly;
    _c->peer             = _peer;
    marcel_mutex_init(&_c->lock, NULL);
  }

  struct circuit_builder_s
  {
    marcel_cond_t wait;
    marcel_mutex_t mutex;
    int rank;
    int completed;
    int openlinks;
  };

  PUK_LIST_TYPE(padico_circuit,
		char*               circuit_id; /**< ID string of the circuit */
		padico_group_t      group;      /**< associated group */
		padico_circuit_cb_t callback;   /**< callback called upon message arrival */
		void*               cb_key;     /**< key associated to the above callback function */
		struct padico_circuit_connection_s*outs;     /**< outgoing connections */
		struct circuit_builder_s builder;            /**< circuit builder associated to this circuit */
		puk_adapter_specific_t as;             /**< adapter-specific data (per-circuit) */
		puk_adapter_vect_t assembly_patterns; /**< set of assembly patterns used by this Circuit */
 		);

#define padico_circuit_send_CHEAPER    1
#define padico_circuit_send_SAFER      2
#define padico_circuit_receive_EXPRESS 3
#define padico_circuit_receive_CHEAPER 4
  typedef int circuit_send_mode_t;
  typedef int circuit_receive_mode_t;

#define circuit_send_CHEAPER    padico_circuit_send_CHEAPER    
#define circuit_send_SAFER      padico_circuit_send_SAFER      
#define circuit_receive_EXPRESS padico_circuit_receive_EXPRESS
#define circuit_receive_CHEAPER padico_circuit_receive_CHEAPER

  struct padico_circuit_driver_s
  {
    /* builder functions */
    void (*create_out)(padico_circuit_t, int);
    void  (*create_in)(padico_circuit_t c, int from);
    padico_rc_t(*getcontact)(padico_circuit_t);
    /* activation */
    void       (*init)(padico_circuit_t);
    void      (*start)(padico_circuit_t);
    void       (*stop)(padico_circuit_t);
    void      (*close)(padico_circuit_t);
    /* pack/unpack functions */
    padico_circuit_connection_t (*begin_packing)(padico_circuit_t, int);
    void          (*pack)(padico_circuit_connection_t, void*, int, 
			  circuit_send_mode_t, circuit_receive_mode_t);
    void   (*end_packing)(padico_circuit_connection_t);
    void        (*unpack)(padico_circuit_connection_t, void*, int,
			  circuit_send_mode_t, circuit_receive_mode_t);
    void (*end_unpacking)(padico_circuit_connection_t);
  };
  PUK_IFACE_TYPE(Circuit, struct padico_circuit_driver_s);

  /** @internal Find the driver for the given connection */
  static inline const struct padico_circuit_driver_s*circuit_get_driver(padico_circuit_connection_t conn)
  {
    const struct padico_circuit_driver_s* driver =
      puk_component_get_driver_Circuit(conn->assembly_pattern, NULL);
    return driver;
  }

  /* *** Circuit management API */

  /** @brief creates a new circuit
   * @param  group_id   The ID of the group on which we create a circuit
   * @param  circuit_id The ID for the new circuit
   * @return a new circuit
   */
  padico_circuit_t padico_circuit_init     (const char*group_id, const char*circuit_id);

  /** @brief Look-up an already created circuit
   * @param  circuit_id the ID of the looked-up circuit
   * @return the circuit, NULL if not found
   */
  padico_circuit_t padico_circuit_lookup   (const char*circuit_id);

  /** @brief hooks to an existing circuit with aditional checks.
   */
  padico_circuit_t padico_circuit_hook(const char*circuit_id, const char*module_id, int header);

  /** @brief activates a handler on an existing circuit
   * @note If a handler is already registered for the circuit,
   * it is replaced with the new one.
   */
  void             padico_circuit_start    (padico_circuit_t, 
					    padico_circuit_cb_t, void*);

  /** @brief deactivate the handler for the given circuit
   */
  void             padico_circuit_shutdown (padico_circuit_t);

  /** @brief destroy a circuit.
   */
  void             padico_circuit_close    (padico_circuit_t c);


  /* ***** Circuit Topology API **************************** */

  /** @brief returns the size (number of nodes) of a circuit
   */
  int              padico_circuit_size     (padico_circuit_t);

  /** @brief returns the rank of the current process (in circuit numbering)
   */
  int              padico_circuit_rank     (padico_circuit_t);

  static inline const char*padico_circuit_getid(padico_circuit_t c)
  {
    return c->circuit_id;
  }

  const char*      padico_circuit_nodename (padico_circuit_t, int);

  padico_string_t  padico_circuit_serialize(padico_circuit_t);


  /* *** Circuit Messaging API */

  static inline padico_circuit_connection_t padico_circuit_begin_packing(padico_circuit_t c,
									 int dest_node)
  {
    padico_circuit_connection_t conn = &c->outs[dest_node];
    marcel_mutex_lock(&conn->lock);
    conn->circuit = c;
    (*(circuit_get_driver(conn)->begin_packing))(c, dest_node);
    return conn;
  }

  static inline void padico_circuit_pack(padico_circuit_connection_t conn,
					 void*ptr, int size,
					 circuit_send_mode_t send_mode, 
					 circuit_receive_mode_t recv_mode)
  {
    if(size > 0)
      (*(circuit_get_driver(conn)->pack))(conn, ptr, size, send_mode, recv_mode);
  }

  static inline void padico_circuit_end_packing(padico_circuit_connection_t conn)
  {
    (*(circuit_get_driver(conn)->end_packing))(conn);
    marcel_mutex_unlock(&conn->lock);
  }

  static inline void padico_circuit_unpack(padico_circuit_connection_t conn,
					   void*ptr, int size,
					   circuit_send_mode_t send_mode, 
					   circuit_receive_mode_t recv_mode)
  {
    if(size > 0)
      (*(circuit_get_driver(conn)->unpack))(conn, ptr, size, send_mode, recv_mode);
  }

  static inline void padico_circuit_end_unpacking(padico_circuit_connection_t conn)
  {
    (*(circuit_get_driver(conn)->end_unpacking))(conn);
  }

  /* *** Circuit Messaging extensions */

#define padico_circuit_begin_packing_deferred(CIRCUIT, DEST) (padico_circuit_begin_packing((CIRCUIT), (DEST)))

  
  /** get the sender number (in circuit numbering system)
   */
  static inline int padico_circuit_get_sender(padico_circuit_connection_t conn)
  {
    return conn->peer;
  }
  /** @} */

#ifndef MAD_LENGTH_ALIGNMENT
#  define MAD_LENGTH_ALIGNMENT sizeof(int)
#  define MAD_ALIGNMENT        sizeof(int)
#  define MAD_ALIGNED          TBX_ALIGN(MAD_ALIGNMENT)
#  define MAD_LENGTH_ALIGNED   TBX_ALIGN(MAD_SIZE_ALIGNMENT)
#endif

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* PADICO_CIRCUIT_H */
