/** @file
 * @brief Padico Circuit Service.
 * @ingroup Circuit
 */

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PadicoTM.h>
#include <Padico/Module.h>
#include "Circuit-internals.h"

PADICO_MODULE_HOOK(Circuit);

static void circuit_builder_create_out(padico_circuit_t c);
static void circuit_builder_master(padico_circuit_t c);
static void circuit_builder_slave(padico_circuit_t c);

void padico_circuit_build_connections(padico_circuit_t c)
{
  c->builder.rank = -1;
  c->builder.completed = 1; /* connected to myself */
  c->builder.openlinks = 0;
  marcel_mutex_init(&c->builder.mutex, NULL);
  marcel_cond_init(&c->builder.wait, NULL);
  if(padico_circuit_rank(c) == 0)
    {
      circuit_builder_master(c);
    }
  else
    {
      circuit_builder_slave(c);
    } 
  marcel_mutex_lock(&c->builder.mutex);
  while(c->builder.completed < padico_circuit_size(c) || ! c->builder.openlinks)
    {
      marcel_cond_wait(&c->builder.wait, &c->builder.mutex);
    }
  marcel_mutex_unlock(&c->builder.mutex);
}

/* on node i, establishes connections to nodes j=i+1..size */
static void circuit_builder_create_out(padico_circuit_t c)
{
  int j;
  for(j = padico_circuit_rank(c) + 1; j < padico_circuit_size(c); j++)
    {
      /* build out from 'rank' to 'j' */
      padico_out(30, "creating out to node %d...\n", j);
      (*(circuit_get_driver(&c->outs[j])->create_out))(c, j);
      marcel_mutex_lock(&c->builder.mutex);
      padico_out(30, "completed out -> #%d\n", j);
      c->builder.completed++;
      marcel_cond_broadcast(&c->builder.wait);
      marcel_mutex_unlock(&c->builder.mutex);
    }
}

/* on node i, establishes in connections from nodes i=0..rank-1 */
static void circuit_builder_create_in(padico_circuit_t c)
{
  int i;
  for(i = 0; i < padico_circuit_rank(c); i++)
    {
      /* receive connections from nodes i=0..rank-1 */
      padico_out(30, "creating in from node %d...\n", i);
      marcel_mutex_lock(&c->builder.mutex);
      c->builder.rank = i;
      (*(circuit_get_driver(&c->outs[i])->create_in))(c, i);
      padico_out(30, "completed in <- #%d\n", i);
      c->builder.completed++;
      marcel_cond_broadcast(&c->builder.wait);
      marcel_mutex_unlock(&c->builder.mutex);
    }
}

static padico_rc_t circuit_builder_getcontact(padico_circuit_t c)
{
  puk_adapter_vect_itor_t z; 
  padico_rc_t rc = padico_rc_msg("<Circuit:nodeContact circuit_id=\"%s\" grank=\"%d\">\n",
				 c->circuit_id, padico_circuit_rank(c));
  puk_vect_foreach(z, puk_adapter, c->assembly_patterns)
    { 
      const struct padico_circuit_driver_s*driver =
	puk_component_get_driver_Circuit(*z, NULL);
      if(driver != NULL && driver->getcontact != NULL)
	{
	  padico_rc_t contact = (*driver->getcontact)(c);
	  if(contact)
	    {
	      padico_rc_add(rc, contact);
	    }
	}
    } 
  padico_rc_add(rc, padico_rc_msg("</Circuit:nodeContact>\n"));
  return rc;
}

static void circuit_builder_master(padico_circuit_t c)
{
  padico_out(30, "building master- circuit_id=**%s**\n", c->circuit_id);
  if(padico_circuit_size(c) > 1)
    {
      /* *1* get the contacts */
      /*    -1.1- locally */
      padico_rc_t lrc = circuit_builder_getcontact(c);
      /* TODO*/
      /*    -1.2- remotely */
      padico_string_t s_getcontact = padico_string_new();
      padico_string_catf(s_getcontact, "<Circuit:getContact circuit_id=\"%s\"/>\n", c->circuit_id);
      padico_group_req_t greq = padico_group_send_to_all(c->group, padico_string_get(s_getcontact));
      padico_rc_t rc = padico_group_wait_all(c->group, greq);
      
      /* *2* build the Circuit:openLinks request */
      padico_string_t s_open = padico_string_new();
      padico_string_catf(s_open, "<Circuit:openLinks circuit_id=\"%s\">\n", c->circuit_id);
      padico_string_catf(s_open, "%s\n", padico_string_get(rc->msg));
      padico_string_catf(s_open, "%s\n", padico_string_get(lrc->msg));
      padico_string_catf(s_open, "</Circuit:openLinks>\n");

      /* *3* steer the link construction */
      /*    -3.1- locally */
      padico_out(30, "parsing request locally\n");
      puk_xml_parse_buffer(padico_string_get(s_open), 
			   strlen(padico_string_get(s_open)), 
			   PUK_TRUST_CONTROL);
      padico_out(30, "local parsing ok.\n");
      /*    -3.2- remotely */
      int i;
      for(i = 1; i < padico_circuit_size(c); i++)
	{
	  padico_out(30, "sending 'openLinks' request to node #%d (%s)\n", i, padico_na_node_getname(padico_group_node(c->group, i)));
	  const padico_na_node_t node = padico_group_node(c->group, i);
	  padico_req_t req = padico_control_send(node, padico_string_get(s_open));
	  padico_out(30, "request to %d sent. Waiting RC...\n", i);
	  padico_tm_req_wait(req);
	  padico_out(30, "request to %d done.\n", i);
	}
      /* *4* clean-up */
      padico_rc_delete(rc);
      padico_string_delete(s_getcontact);
    }
  else
    {
      circuit_builder_create_out(c);
    }
}

static void circuit_builder_slave(padico_circuit_t c)
{
  padico_out(30, "starting slave circuit_id=**%s**\n", c->circuit_id);
  circuit_builder_create_in(c);
}


/* *** Circuit builder: parse ****************************** */


static void circuitgetcontact_start_handler(puk_parse_entity_t e)
{
  const char*circuit_id = puk_parse_getattr(e, "circuit_id");
  padico_rc_t rc = NULL;
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  while(c == NULL)
    {
      padico_out(6, "Circuit:getContact circuit_id=**%s** does not exist. Waiting.\n", circuit_id);
      padico_tm_msleep(200); /* TODO: we are likely to be in a callback. 
				May we actually wait here? */
      c = padico_circuit_lookup(circuit_id);
      if(c != NULL)
	padico_out(6, "Circuit:getContact circuit_id=**%s** created.\n", circuit_id);
    }
  rc = circuit_builder_getcontact(c);
  puk_parse_set_rc(e, rc);
}

static struct puk_tag_action_s circuitgetcontact_action =
{
  .xml_tag        = "Circuit:getContact",
  .start_handler  = &circuitgetcontact_start_handler,
  .end_handler    = NULL,
  .required_level = PUK_TRUST_CONTROL
};


static void circuitnodecontact_start_handler(puk_parse_entity_t e)
{
  const char*circuit_id = puk_parse_getattr(e, "circuit_id");
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  //  int grank = atoi(puk_xml_getattr("grank", xa));
  padico_out(40, "circuit ID=%s\n", circuit_id);
  puk_parse_set_content(e, c);
}
static struct puk_tag_action_s circuitnodecontact_action =
{
  .xml_tag        = "Circuit:nodeContact",
  .start_handler  = &circuitnodecontact_start_handler,
  .end_handler    = NULL,
  .required_level = PUK_TRUST_CONTROL
};

static void circuitopenlinks_start_handler(puk_parse_entity_t e)
{ 
  const char*circuit_id = puk_parse_getattr(e, "circuit_id");
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  padico_out(40, "circuit ID=%s\n", circuit_id);
  puk_parse_set_content(e, c);
}
static void circuitopenlinks_end_handler(puk_parse_entity_t e)
{
  padico_circuit_t c = puk_parse_get_content(e);
  circuit_builder_create_out(c);
  marcel_mutex_lock(&c->builder.mutex);
  c->builder.openlinks = 1;
  marcel_cond_broadcast(&c->builder.wait);
  marcel_mutex_unlock(&c->builder.mutex);
}
static struct puk_tag_action_s circuitopenlinks_action =
{
  .xml_tag        = "Circuit:openLinks",
  .start_handler  = &circuitopenlinks_start_handler,
  .end_handler    = &circuitopenlinks_end_handler,
  .required_level = PUK_TRUST_CONTROL
};

void padico_circuit_builder_init(void)
{
  puk_xml_add_action(circuitgetcontact_action);
  puk_xml_add_action(circuitnodecontact_action);
  puk_xml_add_action(circuitopenlinks_action);
}
