/** @file
 * @brief Padico Circuit/null adapter.
 * @ingroup Circuit
 */

#include <Padico/Module.h>
#include "Circuit.h"

PADICO_MODULE_DECLARE(Circuit_null);
PADICO_MODULE_INIT(padico_circuit_null_init);

static void circuit_null_init (padico_circuit_t);
static void circuit_null_start(padico_circuit_t);
static void circuit_null_stop (padico_circuit_t);
static void circuit_null_close(padico_circuit_t);

static void circuit_null_create_out(padico_circuit_t, int);
static void circuit_null_create_in (padico_circuit_t, int);
static padico_rc_t circuit_null_getcontact(padico_circuit_t);

static padico_circuit_connection_t circuit_null_begin_packing(padico_circuit_t, int);
static void circuit_null_pack(padico_circuit_connection_t, void*, int,
			      circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_null_end_packing(padico_circuit_connection_t);
static void circuit_null_unpack(padico_circuit_connection_t, void*, int,
				circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_null_end_unpacking(padico_circuit_connection_t);


/** @defgroup Circuit_null adapter: Circuit-null- dummy Circuit adapter
 * @note adapter used when a 'Circuit' adapter is required but no communication is expected (e.g. loopback)
 * @ingroup PadicoAdapter
 */

/** The Circuit/SysIO adapter descriptor */
static puk_adapter_t circuit_null_adapter = NULL;

/** 'Circuit' facet for the Circuit/SysIO adapter
 * @ingroup Circuit_null
 */
const static struct padico_circuit_driver_s circuit_null_driver =
  {
    .create_out    = &circuit_null_create_out,
    .create_in     = &circuit_null_create_in,
    .getcontact    = &circuit_null_getcontact,
    .init          = &circuit_null_init,
    .start         = &circuit_null_start,
    .stop          = &circuit_null_stop,
    .close         = &circuit_null_close,
    .begin_packing = &circuit_null_begin_packing,
    .pack          = &circuit_null_pack,
    .end_packing   = &circuit_null_end_packing,
    .unpack        = &circuit_null_unpack,
    .end_unpacking = &circuit_null_end_unpacking
  };


static inline struct circuit_null_in_s* 
circuit_null_inconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_null_in_s*)(conn->adapter_conn_info);
}
static inline struct circuit_null_out_s* 
circuit_null_outconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_null_out_s*)(conn->adapter_conn_info);
}


/* ***** Circuit builder *********************************** */

static void circuit_null_init(padico_circuit_t c)
{
  puk_adapter_setspecific(&c->as, circuit_null_adapter, NULL);
}

static void circuit_null_start(padico_circuit_t c)
{
  /* do nothing */
}

static void circuit_null_stop(padico_circuit_t c)
{
  /* do nothing */
}

static void circuit_null_close(padico_circuit_t c)
{
  /* do nothing */
}

static void circuit_null_create_out(padico_circuit_t c, int dest)
{
  /* do nothing */
}

static void circuit_null_create_in(padico_circuit_t c, int from)
{
  /* do nothing */
}

static padico_rc_t circuit_null_getcontact(padico_circuit_t c)
{
  return NULL;
}

/* ***** Circuit: pack ************************************* */

static padico_circuit_connection_t 
circuit_null_begin_packing(padico_circuit_t circuit,
			    int dest_node)
{
  padico_fatal("Circuit-null: begin_packing() called.\n");
  return NULL;
}


static void circuit_null_pack(padico_circuit_connection_t conn,
			      void*ptr, int size,
			      circuit_send_mode_t send_mode, 
			      circuit_receive_mode_t recv_mode)
{
  padico_fatal("Circuit-null: pack() called.\n");
}


static void circuit_null_end_packing(padico_circuit_connection_t conn)
{
  padico_fatal("Circuit-null: end_packing() called.\n");
}


/* ***** Circuit: unpack *********************************** */

static void circuit_null_unpack(padico_circuit_connection_t conn,
				void*ptr, int size,
				circuit_send_mode_t send_mode, 
				circuit_receive_mode_t recv_mode)
{
  padico_fatal("Circuit-null: unpack() called.\n");
}


static void circuit_null_end_unpacking(padico_circuit_connection_t conn)
{
  padico_fatal("Circuit-null: end_unpacking() called.\n");
}


/* ********************************************************* */

int padico_circuit_null_init(void)
{
  circuit_null_adapter =
    puk_component_declare("Circuit-null",
			  puk_component_provides("Circuit", "circuit", &circuit_null_driver));
  return 0;
}

