/** @file
 * @brief Padico Circuit Service.
 * @ingroup Circuit
 */

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PSP.h>
#include <Padico/PadicoTM.h>
#include <Padico/Module.h>
#include "Circuit-internals.h"

PADICO_MODULE_DECLARE(Circuit_PSP);
PADICO_MODULE_INIT(padico_circuit_psp_init);

static void circuit_psp_init (padico_circuit_t);
static void circuit_psp_start(padico_circuit_t);

static void circuit_psp_conn_init(padico_circuit_t, int);
static padico_rc_t circuit_psp_getcontact(padico_circuit_t);
static void circuit_psp_contact_handler(puk_parse_entity_t);

static padico_circuit_connection_t circuit_psp_begin_packing(padico_circuit_t, int);
static void circuit_psp_pack(padico_circuit_connection_t, void*, int,
			     circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_psp_end_packing(padico_circuit_connection_t);
static void circuit_psp_unpack(padico_circuit_connection_t, void*, int,
			       circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_psp_end_unpacking(padico_circuit_connection_t);

/** @defgroup Circuit_Psp adapter: Circuit/Psp- Circuit over Psp
 * @ingroup PadicoAdapter
 */

/** The Circuit/Psp adapter descriptor */
static puk_adapter_t circuit_psp_adapter = NULL;

/** The Circuit/Psp tag max for this node */
static int max_tag = 1;
static marcel_mutex_t tag_mutex;

/** 'Circuit' facet of the Circuit/Psp adapter 
 * @ingroup Circuit_Psp
 */
const static struct padico_circuit_driver_s circuit_psp_driver =
  {
    .create_out    = &circuit_psp_conn_init,
    .create_in     = &circuit_psp_conn_init,
    .getcontact    = &circuit_psp_getcontact,
    .init          = &circuit_psp_init,
    .start         = &circuit_psp_start,
    .stop          = NULL,
    .close         = NULL,
    .begin_packing = &circuit_psp_begin_packing,
    .pack          = &circuit_psp_pack,
    .end_packing   = &circuit_psp_end_packing,
    .unpack        = &circuit_psp_unpack,
    .end_unpacking = &circuit_psp_end_unpacking
  };

/** Circuit/Psp adapter-specific per-circuit information
 * @ingroup Circuit_Psp
 */
struct circuit_psp_info_s
{
  struct puk_receptacle_PadicoSimplePackets_s psp;
  const char*psp_id;
  size_t h_size;
  int circuit_max_tag;
};

/** Circuit/Psp per-connection information
 * @ingroup Circuit_Psp
 */
struct circuit_psp_out_s
{
  void*sbuf; /**< send buffer from PSP; clear to NULL when used */
  padico_simplepackets_connection_t psp_conn;
};

struct circuit_psp_in_s
{
  const void*header;
  void*token;
  padico_simplepackets_pump_t pump;
};

/** accessor for per circuit adapter-specific info */
static inline struct circuit_psp_info_s*
circuit_psp_specific(padico_circuit_t circuit)
{
  struct circuit_psp_info_s*info =
    puk_adapter_getspecific(&circuit->as, circuit_psp_adapter);
  return info;
}

/** accessors for per connection adapter-specific info (in/out) */
static inline struct circuit_psp_in_s* 
circuit_psp_inconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_psp_in_s*)(conn->adapter_conn_info);
}
static inline struct circuit_psp_out_s* 
circuit_psp_outconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_psp_out_s*)(conn->adapter_conn_info);
}

static void circuit_psp_callback(const void*header,
				 padico_na_node_t from,
				 void*key,
				 padico_simplepackets_pump_t pump,
				 void*token);

/* *** Module init ***************************************** */

static struct puk_tag_action_s circuitpspcontact_action =
{
  .xml_tag        = "Circuit:PSPContact",
  .start_handler  = &circuit_psp_contact_handler,
  .end_handler    = NULL,
  .required_level = PUK_TRUST_CONTROL
};

static int padico_circuit_psp_init(void)
{
  puk_xml_add_action(circuitpspcontact_action);
  circuit_psp_adapter =
    puk_component_declare("Circuit-PSP",
			  puk_component_uses("PadicoSimplePackets", "psp"),
			  puk_component_provides("Circuit", "circuit", &circuit_psp_driver));
  marcel_mutex_init(&tag_mutex, NULL);
  return 0;
}

/* ***** Circuit builder *********************************** */

static void circuit_psp_init(padico_circuit_t c)
{
  padico_string_t s_psp_id = padico_string_new();
  struct circuit_psp_info_s*info = 
    (struct circuit_psp_info_s*)padico_malloc(sizeof(struct circuit_psp_info_s));
  puk_adapter_setspecific(&c->as, circuit_psp_adapter, info);
  info->psp.driver = NULL;
  info->circuit_max_tag = -1;
  
  /* *** begin hack ***************************************** */
  puk_adapter_vect_itor_t j;
  /* Find which assembly we are part of. */
  puk_vect_foreach(j, puk_adapter, c->assembly_patterns)
    {
      const puk_adapter_t a = *j;
      const struct puk_composite_driver_s*const composite_driver = puk_component_get_driver_PadicoComposite(a, NULL);
      if(!composite_driver)
	continue;
      puk_component_conn_vect_itor_t i;
      puk_vect_foreach(i, puk_component_conn, &composite_driver->content->entry_points)
	{
	  if( (i->receptacle->iface == puk_iface_Circuit()) &&
	      (i->context->component == circuit_psp_adapter))
	    {
	      /* We found an assembly whose entry_point is Circuit-PSP.
	       * Hope this is self, instanciate, indirect... and cross fingers!
	       */
	      puk_instance_t global_instance = puk_adapter_instanciate(a);
	      const struct puk_composite_status_s*composite_status = global_instance->status;
	      puk_instance_t local_instance = puk_hashtable_lookup(composite_status->instance_table, i->context);
	      puk_context_indirect_PadicoSimplePackets(local_instance, "psp", &info->psp);
	      goto found;
	    }
	}
    }
 found:
  assert(info->psp.driver != NULL);
  /* *** end hack ******************************************* */
    
  padico_string_catf(s_psp_id, "%s//%s", c->circuit_id, padico_na_node_getname(padico_group_node(c->group, 0)));
  info->psp_id = padico_strdup(padico_string_get(s_psp_id));
  padico_string_delete(s_psp_id);
  padico_out(40, "init Circuit/Psp: %s\n", info->psp_id);
}

static void circuit_psp_start(padico_circuit_t c)
{ 
  struct circuit_psp_info_s*info = circuit_psp_specific(c);
  max_tag = info->circuit_max_tag + 1;
  const uint32_t tag = padico_psp_tag_build(PADICO_PSP_CIRCUIT, max_tag);
  padico_print("Circuit-PSP: starting tag=#%x; id=%s...\n", tag, info->psp_id);
  info->h_size = 0;
  (*info->psp.driver->init)(info->psp._status, tag, info->psp_id, &info->h_size, &circuit_psp_callback, c);
  padico_out(50, "started- h_size=%d\n", info->h_size);
}

static void circuit_psp_conn_init(padico_circuit_t c, int peer_rank)
{
  const struct circuit_psp_info_s*info = circuit_psp_specific(c);
  struct circuit_psp_out_s*outconn = circuit_psp_outconn(&c->outs[peer_rank]);
  outconn->sbuf = NULL;
  if(info->psp.driver->connect)
    {
      const padico_na_node_t node = padico_group_node(c->group, peer_rank);
      (*info->psp.driver->connect)(info->psp._status, node);
    }
}

static padico_rc_t circuit_psp_getcontact(padico_circuit_t c)
{
  padico_rc_t rc;
  int my_max_tag = 0;
  marcel_mutex_lock(&tag_mutex);
  my_max_tag = max_tag++;
  marcel_mutex_unlock(&tag_mutex);
  padico_out(50, "sending contact max_tag=#%x\n", max_tag);
  rc = padico_rc_msg("<Circuit:PSPContact circuit_id=\"%s\" tag=\"%d\"/>\n", c->circuit_id, my_max_tag);
  return rc;
}

static void circuit_psp_contact_handler(puk_parse_entity_t e)
{
  const int tag = atoi(puk_parse_getattr(e, "tag"));
  const char*circuit_id = puk_parse_getattr(e, "circuit_id");
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  padico_out(50, "received contact tag=#%x\n", tag);
  marcel_mutex_lock(&tag_mutex);
  if(tag > circuit_psp_specific(c)->circuit_max_tag)
    circuit_psp_specific(c)->circuit_max_tag = tag;
  marcel_mutex_unlock(&tag_mutex);
}

/* ***** Circuit: pack ************************************* */

static padico_circuit_connection_t 
circuit_psp_begin_packing(padico_circuit_t circuit, int dest_rank)
{
  const padico_na_node_t dest_node = padico_group_node(circuit->group, dest_rank);
  padico_circuit_connection_t circuit_conn = &circuit->outs[dest_rank];
  struct circuit_psp_info_s*const info = circuit_psp_specific(circuit);
  struct circuit_psp_out_s*const outconn = circuit_psp_outconn(circuit_conn);
  outconn->psp_conn  = (*info->psp.driver->new_message)(info->psp._status, dest_node, &outconn->sbuf);
  return circuit_conn;
}

static void circuit_psp_pack(padico_circuit_connection_t _conn,
			     void*ptr, int size,
			     circuit_send_mode_t send_mode, 
			     circuit_receive_mode_t recv_mode)
{
  struct circuit_psp_out_s*const outconn = circuit_psp_outconn(_conn);
  struct circuit_psp_info_s*const info   = circuit_psp_specific(_conn->circuit);
  if(outconn->sbuf != NULL)
    {
      if(size <= info->h_size)
	{
	  memcpy(outconn->sbuf, ptr, size);
	} 
      else
	{
	  memcpy(outconn->sbuf, ptr, info->h_size);
	  (*info->psp.driver->pack)(info->psp._status, outconn->psp_conn, ptr + info->h_size, size - info->h_size);
	}
      outconn->sbuf = NULL;
    }
  else
    {
      (info->psp.driver->pack)(info->psp._status, outconn->psp_conn, ptr, size);
    }
}

static void circuit_psp_end_packing(padico_circuit_connection_t _conn)
{
  struct circuit_psp_out_s*const outconn = circuit_psp_outconn(_conn);
  struct circuit_psp_info_s*const info   = circuit_psp_specific(_conn->circuit);
  (*info->psp.driver->end_message)(info->psp._status, outconn->psp_conn);
  outconn->sbuf = NULL;
}

/* ***** Circuit: unpack *********************************** */

static void circuit_psp_unpack(padico_circuit_connection_t _conn,
			       void*ptr, int size,
			       circuit_send_mode_t send_mode, 
			       circuit_receive_mode_t recv_mode)
{
  struct circuit_psp_in_s*const inconn = circuit_psp_inconn(_conn);
  struct circuit_psp_info_s*const info = circuit_psp_specific(_conn->circuit);
  if(inconn->header == NULL)
    {
      /* regular unpack */
      (*inconn->pump)(inconn->token, ptr, size);
    }
   else
    {
      /* first unpack */
      if(size <= info->h_size)
	{
	  /* pre-buffered data in header */
	  memcpy(ptr, inconn->header, size);
	}
      else
	{
	  /* first pack didn't fit in header */
	  memcpy(ptr, inconn->header, info->h_size);
	  (*inconn->pump)(inconn->token, ptr + info->h_size, size - info->h_size);
	}
      inconn->header = NULL;
    }
}

static void circuit_psp_end_unpacking(padico_circuit_connection_t _conn)
{
  struct circuit_psp_in_s*const inconn = circuit_psp_inconn(_conn);
  inconn->header = NULL;
  inconn->pump   = NULL;
  inconn->token  = NULL;
}

/* ********************************************************* */

static void circuit_psp_callback(const void*header,
				 padico_na_node_t from,
				 void*key,
				 padico_simplepackets_pump_t pump,
				 void*token)
{
  const padico_circuit_t circuit = (padico_circuit_t)key;
  int rank = -1;
  struct padico_circuit_connection_s circuit_conn;
  struct circuit_psp_in_s*const inconn = circuit_psp_inconn(&circuit_conn);
  int i;
  for(i = 0; i < padico_group_size(circuit->group); i++)
    {
      if(padico_group_node(circuit->group, i) == from)
	{
	  rank = i;
	  break;
	}
    }
  assert(rank != -1);
  padico_circuit_connection_init(&circuit_conn, circuit, 
				 circuit->outs[rank].assembly_pattern, 
				 /* take the assembly info from the out connection */
				 rank);
  inconn->header  = header;
  inconn->pump    = pump;
  inconn->token   = token;
  (*circuit->callback)(&circuit_conn, circuit->cb_key);
}
