/** @file
 * @brief Padico Circuit Service.
 * @ingroup Circuit
 */

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PadicoTM.h>
#include <Padico/Module.h>
#include <Padico/SysIO.h>
#include "Circuit.h"

PADICO_MODULE_DECLARE(Circuit_SysIO);
PADICO_MODULE_INIT(padico_circuit_sysio_init);

static void circuit_sysio_init (padico_circuit_t);
static void circuit_sysio_start(padico_circuit_t);
static void circuit_sysio_stop (padico_circuit_t);
static void circuit_sysio_close(padico_circuit_t);

static void circuit_sysio_create_out(padico_circuit_t, int);
static void circuit_sysio_create_in (padico_circuit_t, int);
static padico_rc_t circuit_sysio_getcontact(padico_circuit_t);

static padico_circuit_connection_t circuit_sysio_begin_packing(padico_circuit_t, int);
static void circuit_sysio_pack(padico_circuit_connection_t, void*, int,
			       circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_sysio_end_packing(padico_circuit_connection_t);
static void circuit_sysio_unpack(padico_circuit_connection_t, void*, int,
				 circuit_send_mode_t, circuit_receive_mode_t);
static void circuit_sysio_end_unpacking(padico_circuit_connection_t);


/** @defgroup Circuit_SysIO adapter: Circuit/SysIO- Circuit over system I/O
 * @ingroup PadicoAdapter
 */

/** The Circuit/SysIO adapter descriptor */
static puk_adapter_t circuit_sysio_adapter = NULL;

/** 'Circuit' facet of the Circuit/SysIO adapter
 * @ingroup Circuit_SysIO
 */
const static struct padico_circuit_driver_s circuit_sysio_driver =
  {
    .create_out    = &circuit_sysio_create_out,
    .create_in     = &circuit_sysio_create_in,
    .getcontact    = &circuit_sysio_getcontact,
    .init          = &circuit_sysio_init,
    .start         = &circuit_sysio_start,
    .stop          = &circuit_sysio_stop,
    .close         = &circuit_sysio_close,
    .begin_packing = &circuit_sysio_begin_packing,
    .pack          = &circuit_sysio_pack,
    .end_packing   = &circuit_sysio_end_packing,
    .unpack        = &circuit_sysio_unpack,
    .end_unpacking = &circuit_sysio_end_unpacking
  };

/** Circuit/SysIO adapter-specific per-circuit information block
 * @ingroup Circuit_SysIO
 */
struct circuit_sysio_info_s
{
  padico_io_t*in_ios;      /**< input I/O handles */
  padico_io_t builder_io;  /**< listener sys I/O endpoint */
  int*out_ports;           /**< other nodes ports vector for constructing the topology */
};
/** Circuit/SysIO per-connection information blocks */
struct circuit_sysio_out_s
{
  int out_fd;
};
struct circuit_sysio_in_s
{
  int in_fd;
};

/** accessor for per circuit adapter-specific info */
static inline struct circuit_sysio_info_s*
circuit_sysio_specific(padico_circuit_t circuit)
{
 struct circuit_sysio_info_s*info=
   puk_adapter_getspecific(&circuit->as, circuit_sysio_adapter);
 return info;
}

static inline struct circuit_sysio_in_s* 
circuit_sysio_inconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_sysio_in_s*)(conn->adapter_conn_info);
}
static inline struct circuit_sysio_out_s* 
circuit_sysio_outconn(struct padico_circuit_connection_s*conn)
{
  return (struct circuit_sysio_out_s*)(conn->adapter_conn_info);
}

static int circuit_builder_sysio_acceptor(int fd, void*key);
static int circuit_sysio_callback_reader(int fd, void*key); 
static void circuitsysiocontact_start_handler(puk_parse_entity_t e);

/* *** Module init ***************************************** */

static struct puk_tag_action_s circuitsysiocontact_action =
{
  .xml_tag        = "Circuit:SysIOContact",
  .start_handler  = &circuitsysiocontact_start_handler,
  .end_handler    = NULL,
  .required_level = PUK_TRUST_CONTROL
};

static int padico_circuit_sysio_init(void)
{
  puk_xml_add_action(circuitsysiocontact_action);
  circuit_sysio_adapter =
    puk_component_declare("Circuit-SysIO",
			  puk_component_provides("Circuit", "circuit", &circuit_sysio_driver));
  return 0;
}

/* ***** Circuit builder *********************************** */

static void circuit_sysio_init(padico_circuit_t c)
{
  struct circuit_sysio_info_s*info = 
    (struct circuit_sysio_info_s*)padico_malloc(sizeof(struct circuit_sysio_info_s));
  padico_out(40, "init Circuit/SysIO\n");
  puk_adapter_setspecific(&c->as, circuit_sysio_adapter, info);
  info->builder_io = NULL;
  info->out_ports = (int*)padico_malloc(sizeof(int)*padico_circuit_size(c));
  memset(info->out_ports, 0, sizeof(int)*padico_circuit_size(c));
  info->in_ios = (padico_io_t*)padico_calloc(padico_circuit_size(c), sizeof(padico_io_t));
}

static void circuit_sysio_start(padico_circuit_t c)
{
  int i;
  for(i=0; i<padico_circuit_size(c); i++)
    {
      if(circuit_sysio_specific(c)->in_ios != NULL && 
	 circuit_sysio_specific(c)->in_ios[i] != NULL)
	{
	  padico_io_activate(circuit_sysio_specific(c)->in_ios[i]);
	}
    }
}

static void circuit_sysio_stop(padico_circuit_t c)
{
  int i;
  for(i=0; i<padico_circuit_size(c); i++)
    {
      if(circuit_sysio_specific(c)->in_ios[i] != NULL)
	{
	  const int fd = padico_sysio_getfd(circuit_sysio_specific(c)->in_ios[i]);
	  padico_sysio_shutdown_wr(fd);
	  padico_io_deactivate(circuit_sysio_specific(c)->in_ios[i]);
	}
    }
}

static void circuit_sysio_close(padico_circuit_t c)
{
  int i;
  for(i=0; i<padico_circuit_size(c); i++)
    {
      if(circuit_sysio_specific(c)->in_ios[i] != NULL)
	{
	  const int in_fd = padico_sysio_getfd(circuit_sysio_specific(c)->in_ios[i]);
	  padico_io_unregister(circuit_sysio_specific(c)->in_ios[i]);
	  padico_sysio_close(in_fd);
	}
    }
  padico_free(circuit_sysio_specific(c)->in_ios);
  padico_free(circuit_sysio_specific(c)->out_ports);
}

static void circuit_sysio_create_out(padico_circuit_t c, int dest)
{
  padico_na_node_t node = padico_group_node(c->group, dest);
  padico_na_host_t host = padico_na_node_gethost(node);
  const struct in_addr*addr   = padico_na_host_getaddr(host);
  padico_out(30, "use SysIO for out to node=%d (hostname=%s)- waiting ports...\n", 
	     dest, padico_na_host_getname(host));
  marcel_mutex_lock(&c->builder.mutex);
  while((circuit_sysio_specific(c)->out_ports == NULL) || 
	(circuit_sysio_specific(c)->out_ports[dest] == 0))
    {
      marcel_cond_wait(&c->builder.wait, &c->builder.mutex);
    }
  padico_out(30, "out to node=%d-- building socket\n", dest);
  circuit_sysio_outconn(&c->outs[dest])->out_fd =
    padico_sysio_connected_inet_socket(addr, circuit_sysio_specific(c)->out_ports[dest]);
  circuit_sysio_specific(c)->in_ios[dest] = 
    padico_io_register(circuit_sysio_outconn(&c->outs[dest])->out_fd,
		       PADICO_IO_EVENT_READ, 
		       &circuit_sysio_callback_reader, &c->outs[dest]);
  marcel_mutex_unlock(&c->builder.mutex);
  padico_out(30, "out to node=%d-- socket fd=%d\n", 
	     dest, circuit_sysio_outconn(&c->outs[dest])->out_fd);
}

static void circuit_sysio_create_in(padico_circuit_t c, int from)
{
  /* create sysio_info.builder_io */
  padico_out(50, "waiting builder IO...\n");
  while(circuit_sysio_specific(c)->builder_io == NULL)
    {
      marcel_cond_wait(&c->builder.wait, &c->builder.mutex);
    }
  padico_out(50, "builder IO ok.\n");

  padico_out(30, "use SysIO for in from node=%d -- waiting in_ios[]...\n", from);
  padico_io_activate(circuit_sysio_specific(c)->builder_io);
  while(circuit_sysio_specific(c)->in_ios[from] == NULL)
    {
      marcel_cond_wait(&c->builder.wait, &c->builder.mutex);
    }
  padico_out(30, "in from node=%d ready (fd=%d).\n",
	     from, padico_sysio_getfd(circuit_sysio_specific(c)->in_ios[from]));
}

static padico_rc_t circuit_sysio_getcontact(padico_circuit_t c)
{
  int port = 0;
  int fd = padico_sysio_bound_inet_socket(&port);
  padico_rc_t rc;
  marcel_mutex_lock(&c->builder.mutex);
  circuit_sysio_specific(c)->builder_io =
    padico_io_register(fd, PADICO_IO_EVENT_READ,
		       &circuit_builder_sysio_acceptor, c);
  marcel_cond_broadcast(&c->builder.wait);
  marcel_mutex_unlock(&c->builder.mutex);
  rc = padico_rc_msg("<Circuit:SysIOContact circuit_id=\"%s\" grank=\"%d\" port=\"%d\"/>\n",
		     c->circuit_id, padico_circuit_rank(c), port);
  return rc;
}

static void circuitsysiocontact_start_handler(puk_parse_entity_t e)
{
  int port = atoi(puk_parse_getattr(e, "port"));
  const char*circuit_id = puk_parse_getattr(e, "circuit_id");
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  int grank = atoi(puk_parse_getattr(e, "grank"));
  marcel_mutex_lock(&c->builder.mutex);
  circuit_sysio_specific(c)->out_ports[grank] = port;
  marcel_mutex_unlock(&c->builder.mutex);
}

/* ***** Circuit: pack ************************************* */

static padico_circuit_connection_t 
circuit_sysio_begin_packing(padico_circuit_t circuit,
			    int dest_node)
{
  padico_circuit_connection_t circuit_conn = &circuit->outs[dest_node];
  return circuit_conn;
}


static void circuit_sysio_pack(padico_circuit_connection_t conn,
			       void*ptr, int size,
			       circuit_send_mode_t send_mode, 
			       circuit_receive_mode_t recv_mode)
{
  int rc = padico_sysio_out(circuit_sysio_outconn(conn)->out_fd, ptr, size);
  if(rc != size)
    padico_warning("Circuit-SysIO: trying to send data on closed channel.\n");
}


static void circuit_sysio_end_packing(padico_circuit_connection_t conn)
{
  /* empty */
}


/* ***** Circuit: unpack *********************************** */

static void circuit_sysio_unpack(padico_circuit_connection_t conn,
				 void*ptr, int size,
				 circuit_send_mode_t send_mode, 
				 circuit_receive_mode_t recv_mode)
{
  int rc = padico_sysio_in(circuit_sysio_inconn(conn)->in_fd, ptr, size);
  if(rc != size)
    padico_warning("Circuit-SysIO: missing data for unpack- size=%d; rc=%d\n", size, rc);
}


static void circuit_sysio_end_unpacking(padico_circuit_connection_t conn)
{
  /* empty */
}


/* ********************************************************* */

static void*circuit_builder_sysio_worker(void*arg)
{
  padico_circuit_t c = (padico_circuit_t)arg;
  int peer = c->builder.rank;
  marcel_detach(marcel_self());
  padico_out(30, "register in_ios[%d]\n", peer);
  marcel_mutex_lock(&c->builder.mutex);
  circuit_sysio_specific(c)->in_ios[c->builder.rank] = 
    padico_io_register(circuit_sysio_outconn(&c->outs[peer])->out_fd,
		       PADICO_IO_EVENT_READ, 
		       &circuit_sysio_callback_reader, &c->outs[peer]);
  assert(circuit_sysio_specific(c)->in_ios[peer] != NULL);
  marcel_cond_broadcast(&c->builder.wait);
  marcel_mutex_unlock(&c->builder.mutex);
  return NULL;
}

static int circuit_builder_sysio_acceptor(int fd, void*key)
{
  int again = 0;
  padico_circuit_t c = (padico_circuit_t)key;
  struct sockaddr_in addr;
  socklen_t addr_len = sizeof(addr);
  int conn = -1;
  padico_out(50, "entering sysio_io acceptor\n");
  conn = padico_sysio_accept(fd, (struct sockaddr*)&addr, &addr_len);
  marcel_mutex_lock(&c->builder.mutex);
  circuit_sysio_outconn(&c->outs[c->builder.rank])->out_fd = conn;
  marcel_mutex_unlock(&c->builder.mutex);
  padico_tm_invoke_later(&circuit_builder_sysio_worker, key);
  padico_out(50, "exiting sysio_io acceptor\n");
  return again;
}

static int circuit_sysio_callback_reader(int fd, void*key)
{
  int again = 1;
  if(padico_sysio_eos(fd))
    {
      again = 0;
    }
  else
    {
      const struct padico_circuit_connection_s*out_conn = key;
      const padico_circuit_t circuit = out_conn->circuit;
      struct padico_circuit_connection_s conn;
      padico_circuit_connection_init(&conn, circuit, out_conn->assembly_pattern, out_conn->peer);
      circuit_sysio_inconn(&conn)->in_fd = fd;
      (*circuit->callback)(&conn, circuit->cb_key);
    }
  return again;
}


