/** @file
 * @brief Secondary connector- provides PadicoConnector interface
 * atop PadicoBootstrap to enable the use of primary controlers as secondary.
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include <Padico/Topology.h>
#include <Padico/PadicoControl.h>
#include <Padico/AddrDB.h>


PADICO_MODULE_DECLARE(SecondaryConnector);
PADICO_MODULE_INIT(padico_secondary_connector_init);

/* ********************************************************* */

static void*secondary_connector_instanciate(puk_instance_t instance, puk_context_t context);
static void secondary_connector_destroy(void*_status);

const static struct puk_adapter_driver_s secondary_connector_adapter_driver =
{
  .instanciate = &secondary_connector_instanciate,
  .destroy     = &secondary_connector_destroy
};

static int secondary_connector_connect(void*_status, padico_req_t req, padico_na_node_t node);

const static struct padico_control_connector_driver_s secondary_connector_driver =
{
  .connect = &secondary_connector_connect
};

struct secondary_connector_status_s
{
  struct puk_receptacle_PadicoBootstrap_s bootstrap;
};

/* ********************************************************* */

static int padico_secondary_connector_init(void)
{
  puk_component_declare("SecondaryConnector",
			puk_component_provides("PadicoAdapter",   "adapter",   &secondary_connector_adapter_driver),
			puk_component_provides("PadicoConnector", "connector", &secondary_connector_driver),
			puk_component_uses("PadicoBootstrap", "bootstrap"));
  return 0;
}

/* ********************************************************* */


static void*secondary_connector_instanciate(puk_instance_t instance, puk_context_t context)
{
  struct secondary_connector_status_s*status = padico_malloc(sizeof(struct secondary_connector_status_s));
  puk_context_indirect_PadicoBootstrap(instance, "bootstrap", &status->bootstrap);
  return status;
}
static void secondary_connector_destroy(void*_status)
{
  padico_free(_status);
}
static int secondary_connector_connect(void*_status, padico_req_t req, padico_na_node_t node)
{
  padico_print("SecondaryConnector: connect- local=%s; remote=%s\n",
	       (char*)padico_na_node_getuuid(padico_na_getlocalnode()),
	       (char*)padico_na_node_getuuid(node));
  struct secondary_connector_status_s*status = _status;
  int rc = 0;
  const int is_server = (memcmp(padico_na_node_getuuid(node), 
				padico_na_node_getuuid(padico_na_getlocalnode()),
				PADICO_TOPO_UUID_SIZE) > 0);
  const static char component[] = "SecondaryConnector";
  const static char singleton_name[] = "secondary-connector";
  const int singleton_name_size = strlen(singleton_name);
  struct padico_bootstrap_addr_s addr = { .ip.s_addr = htonl(INADDR_ANY), .port = htons(0) };
  if(is_server)
    {
      padico_print("SecondaryConnector: bootstrap_new\n");
      rc = status->bootstrap.driver->bootstrap_new(status->bootstrap._status, &addr.ip, &addr.port);
      padico_addrdb_publish(node, component, singleton_name, singleton_name_size, &addr, sizeof(addr));
      padico_print("SecondaryConnector: bootstrap_new done.\n");
    }
  else
    {
      padico_print("SecondaryConnector: waiting address for bootstrap_connect\n");
      padico_req_t req = padico_tm_req_new(NULL, NULL);
      padico_addrdb_get(node, component, singleton_name, singleton_name_size, &addr, sizeof(addr), req);
      padico_tm_req_wait(req);
      padico_print("SecondaryConnector: bootstrap_connect\n");
      rc = status->bootstrap.driver->bootstrap_connect(status->bootstrap._status, &addr.ip, addr.port);
      padico_print("SecondaryConnector: bootstrap_connect- done\n");
    }
  return rc;
}


