/** @file
 * @ingroup Control
 */


#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include <Padico/NetSelector.h>
#include <Padico/PadicoControl.h>

PADICO_MODULE_DECLARE(ControlSimpleSelector);
PADICO_MODULE_INIT(control_simpleselector_init);
PADICO_MODULE_FINALIZE(control_simpleselector_finalize);

/** @defgroup ControlSimpleSelector adapter: ControlSimpoleSelector- simple selector for the control channel
 * @ingroup PadicoAdapter
 */

static struct
{
  puk_hashtable_t table; /* hash: node -> controler instance */
  puk_component_t selector;
} cs_selector = { .selector = NULL };

static void css_event_listener(void*arg);

static puk_instance_t control_select_controler(padico_na_node_t node);

/* ********************************************************* */

/** 'ControlSelector' facet for the ControlSimpleSelector adapter
 * @ingroup ControlSimpleSelector
 */
const static struct padico_control_selector_driver_s control_simpleselector_driver =
  {
    .router  = &control_select_controler
  };

/* ********************************************************* */

static int control_simpleselector_init(void)
{
  cs_selector.selector =
    puk_component_declare("ControlSimpleSelector", 
			  puk_component_provides("ControlSelector", "selector", &control_simpleselector_driver));
  cs_selector.table = puk_hashtable_new_ptr();
  padico_control_event_subscribe(&css_event_listener);
  padico_control_plug_selector(cs_selector.selector);
  return 0;
}

static void control_simpleselector_finalize(void)
{
  padico_control_event_disable();
  padico_control_event_flush();
  padico_control_unplug_selector(cs_selector.selector);
  cs_selector.selector = NULL;
  padico_control_event_unsubscribe(&css_event_listener);
  padico_control_event_flush();
  padico_control_event_enable();
}

/* ********************************************************* */
/* *** ControlSimpleSelector ******************************* */

/** @pre event channel "Control" disabled
 */
static puk_instance_t control_select_controler(padico_na_node_t node)
{
  padico_out(40, "node=%s\n", padico_na_node_getname(node));
  padico_control_event_disable();
  padico_control_event_flush();
  const puk_instance_t instance = puk_hashtable_lookup(cs_selector.table, node);
  padico_control_event_enable();
  return instance;
}

/* ********************************************************* */

static void css_event_listener(void*arg)
{
  padico_control_event_t event = (struct padico_control_event_s*)arg;
  switch(event->kind)
    {
    case PADICO_CONTROL_EVENT_INIT_CONTROLER:
      {
	const padico_na_node_vect_t nodes = event->INIT_CONTROLER.nodes;
	const puk_instance_t controler = event->INIT_CONTROLER.controler;
	padico_out(40, "init node event- %d nodes; controler=%s\n",
		   padico_na_node_vect_size(nodes), controler->component->name);
	padico_na_node_vect_itor_t i;
	puk_vect_foreach(i, padico_na_node, nodes)
	  {
	    const padico_na_node_t node = *i;
	    puk_hashtable_insert(cs_selector.table, node, controler);
	    padico_out(40, "init node event- new node *%s* served by %s\n",
		       padico_na_node_getname(node), controler->component->name);
	  }
      }
      break;

    case PADICO_CONTROL_EVENT_NEW_NODE:
      {
	const padico_na_node_t node = event->NODE.node;
	const puk_instance_t controler = event->NODE.controler;
	assert(controler != NULL);
	if(puk_hashtable_lookup(cs_selector.table, node) == NULL)
	  {
	    padico_out(40, "node event- new node *%s* served by %s\n",
		       padico_na_node_getname(node), controler->component->name);
	    puk_hashtable_insert(cs_selector.table, node, controler);
	  }
      }
      break;

    case PADICO_CONTROL_EVENT_DELETE_NODE:
      {
	const padico_na_node_t node = event->NODE.node;
	if(puk_hashtable_lookup(cs_selector.table, node) != NULL)
	  {
	    puk_hashtable_remove(cs_selector.table, node);
	  }
      }
      break;

    default:
      break;
    }
}
