/** @file
 * @brief LZO Packet Filter adapter
 */

#include <Padico/Module.h>
#include "PacketFilter.h"
#include "minilzo.h"

PADICO_MODULE_DECLARE(PacketFilter_LZO);
PADICO_MODULE_INIT(lzopf_init);

/* ********************************************************* */
/* *** LZO Packet Filter *********************************** */
/* ********************************************************* */


/** @defgroup PacketFilter_LZO adapter: PacketFilter-LZO- a packet filter that performs realtime LZO compression
 * @ingroup PadicoAdapter
 */

static void* lzopf_instanciate(puk_instance_t assembly, puk_context_t context);
static void lzopf_destroy(void*state);
static void lzopf_encode(void*, const void*, size_t, void**, size_t*);
static void lzopf_decode(void*, const void*, size_t, void*, size_t*);

/** instanciation facet for PacketFilter-LZO
 * @ingroup PacketFilter_LZO
 */
const static struct puk_adapter_driver_s lzo_adapter_driver =
  {
    .instanciate = &lzopf_instanciate,
    .destroy     = &lzopf_destroy
  };
/** 'PacketFilter' facet for PacketFilter-LZO
 * @ingroup PacketFilter_LZO
 */
const static struct packetfilter_driver_s lzopf_driver = 
  {
    .pf_encode = &lzopf_encode,
    .pf_decode = &lzopf_decode
  };

struct lzo_state_s
{
  void*wrkmem;
};

static void* lzopf_instanciate(puk_instance_t assembly, puk_context_t context)
{
  struct lzo_state_s*state = padico_malloc(sizeof(struct lzo_state_s));
  state->wrkmem = padico_malloc(LZO1X_MEM_COMPRESS);
  return state;
}

static void lzopf_destroy(void*w)
{
  struct lzo_state_s*state  = w;
  assert(state != NULL);
  padico_free(state->wrkmem);
  padico_free(state);
}


/* 'out' allocated by encode function
 */
static void lzopf_encode(void*w,
			 const void*in, size_t in_size,
			 void**out, size_t* out_size)
{
  struct lzo_state_s*state = w;
  lzo_uint s = in_size + in_size / 16 + 64;
  void*p = padico_malloc(s);
  int rc = lzo1x_1_compress(in, in_size, p, &s, state->wrkmem);
  if(rc != LZO_E_OK)
    {
      /* compression failed: don't compress */
      padico_fatal("LZO: compression failed.\n");
      memcpy(p, in, in_size);
      s = in_size;
    }
  *out = p;
  *out_size = s;
}

/* 'out' allocated by caller
*/
static void lzopf_decode(void*w,
			 const void*in, size_t in_size,
			 void*out, size_t* out_size)
{
  lzo_uint s = *out_size;
  int rc = lzo1x_decompress(in, in_size, out, &s, NULL);
  if(rc != LZO_E_OK)
    {
      padico_fatal("LZO: CRC error while uncompressing- in_size=%lu; out_size=%lu; rc=%d\n",
		   (unsigned long)in_size, (unsigned long)*out_size, rc);
    }
  *out_size = s;
}


static int lzopf_init(void)
{
  lzo_init();
  padico_print("PacketFilter-LZO: using miniLZO version %s (%s)\n",
	       lzo_version_string(), lzo_version_date());

  puk_component_declare("PacketFilter-LZO",
			puk_component_provides("PadicoAdapter", "adapter", &lzo_adapter_driver),
			puk_component_provides("PacketFilter", "pf", &lzopf_driver));
  return 0;
}
