/** @file NetAccess remote control -- control channel management
 * @ingroup NetAccess
 */

#ifndef PADICO_REMOTECONTROL_H
#define PADICO_REMOTECONTROL_H

#include <Padico/Puk.h>
#include <Padico/PadicoTM.h>
#include <Padico/Topology.h>


#ifdef __cplusplus
extern "C" {
#endif

  /** @defgroup NetAccessControl API: NetAccess control channel
   * @ingroup NetAccess
   */

  /** @addtogroup NetAccessControl
   * @{
   */

  /* *** Internal PadicoControl API */
  
  PUK_VECT_TYPE(padico_string, padico_string_t);

  /** A control message.
   * Message is built in vect 'content':
   *   first vect entry is XML message; following entries are binary attachments.
   * Ready to send on network is contained in iovec 'v'.
   */
  struct padico_control_msg_s
  {
    struct padico_string_vect_s content;
    struct iovec*v;
    size_t size;
    int offset; /**< cumulated size of parts */
  };
  typedef struct padico_control_msg_s*padico_control_msg_t;

  /** create a new empty control message */
  static inline padico_control_msg_t padico_control_msg_new(void)
  {
    padico_control_msg_t msg = padico_malloc(sizeof(struct padico_control_msg_s));
    padico_string_vect_init(&msg->content);
    padico_string_t v0 = padico_string_new();
    padico_string_vect_push_back(&msg->content, v0);
    msg->v = NULL;
    msg->size = 0;
    msg->offset = 0;
    return msg;
  }
  /** destroy a control message */
  static inline void padico_control_msg_destroy(padico_control_msg_t msg)
  {
    padico_string_vect_itor_t i;
    puk_vect_foreach(i, padico_string, &msg->content)
      {
	padico_string_delete(*i);
      }
    padico_string_vect_destroy(&msg->content);
    if(msg->v != NULL)
      {
	padico_free(msg->v);
	msg->v = NULL;
      }
    padico_free(msg);
  }

#define padico_control_msg_add_cmd(MSG, TEXT, ...)				\
  padico_string_catf(padico_string_vect_at(&(MSG)->content, 0), (TEXT), ##__VA_ARGS__ )
    
  /** add an attached binary part to a control message */
  static inline void padico_control_msg_add_part(padico_control_msg_t msg, const void*text, size_t len)
  {
    padico_string_t part = padico_string_new();
    padico_string_resize(part, len);
    memcpy(padico_string_get(part), text, len);
    padico_string_vect_push_back(&msg->content, part);
    padico_control_msg_add_cmd(msg, "<puk:part offset=\"%d\" len =\"%d\"/>", msg->offset, len);
    msg->offset += len;
  }

  static inline const char*padico_control_msg_get_cmd(padico_control_msg_t msg)
  {
    padico_string_t first = padico_string_vect_at(&msg->content, 0);
    return padico_string_get(first);
  }

  static inline int padico_control_msg_get_iovec_size(padico_control_msg_t msg)
  {
    assert(msg->v != NULL);
    return padico_string_vect_size(&msg->content);
  }

  static inline const struct iovec*padico_control_msg_get_iovec(padico_control_msg_t msg)
  {
    const int n = padico_string_vect_size(&msg->content);
    struct iovec*v = padico_malloc(sizeof(struct iovec) * (n + 1));
    int i;
    for(i = 0; i < n; i++)
      {
	padico_string_t s = padico_string_vect_at(&msg->content, i);
	v[i].iov_base = padico_string_get(s);
	v[i].iov_len = padico_string_size(s);
	msg->size += padico_string_size(s);
      }
    v[n].iov_base = NULL;
    v[n].iov_len = -1;
    msg->v = v;
    return v;
  }

  static inline size_t padico_control_msg_get_size(padico_control_msg_t msg)
  {
    assert(msg->v != NULL);
    return msg->size;
  }

  static inline struct iovec padico_control_msg_get_bytes(padico_control_msg_t msg)
  {
      const struct iovec*v = padico_control_msg_get_iovec(msg);
      const size_t size = padico_control_msg_get_size(msg);
      char*bytes = padico_malloc(size);
      int offset = 0;
      while(v->iov_base != NULL)
	{
	  memcpy(bytes + offset, v->iov_base, v->iov_len);
	  offset += v->iov_len;
	  v++;
	}
      return (struct iovec){ bytes, size };
  }

  /** Header for PadicoControl messages, in NET byte order, supposed to be 32-bit-aligned.
   */
  struct padico_control_packed_hdr_s
  {
    uint32_t _req_id; /**< request ID (in the orignating node's namespace) */
    uint32_t _size;   /**< size of the following command string */
  };
  struct padico_control_host_hdr_s
  {
    uint32_t req_id;
    size_t size;
  };

  static inline void padico_control_pack_hdr(struct padico_control_packed_hdr_s*packed_hdr,
					     const struct padico_control_host_hdr_s*host_hdr)
  {
    packed_hdr->_req_id = htonl(host_hdr->req_id);
    packed_hdr->_size   = htonl(((uint32_t)host_hdr->size));
  };

  static inline void padico_control_unpack_hdr(struct padico_control_host_hdr_s*host_hdr,
					       const struct padico_control_packed_hdr_s*packed_hdr)
  {
    host_hdr->req_id = htonl(packed_hdr->_req_id);
    host_hdr->size   = htonl(((uint32_t)packed_hdr->_size));
  };


  /* The home interface for PadicoControl */

  /** @internal */
  padico_tasklet_t padico_control_get_event_channel(void);
  /** @internal */
  void             padico_control_gen_init_events(padico_event_listener_t l);

  /** deliver a message from a Controler */
  void padico_control_deliver_message(puk_instance_t controler, padico_na_node_t from, char*cmd, size_t size);

  void padico_control_send_ext(padico_na_node_t node, const char*cmd, uint32_t req_id);

  /** the real function to send a message on control channel- most of the time, we use a wraper, however. */
  int  padico_control_send_msg(padico_na_node_t node, padico_control_msg_t msg);

#ifndef __cplusplus
  
  /** Send a "one way" message on the control channel. */
  static inline void padico_control_send_oneway(padico_na_node_t node, const char*msg)
  {
    padico_control_send_ext(node, msg, PADICO_REQ_ID_NONE);
  }

  /** Send an async request on the control channel. */
  static inline void padico_control_send_async(padico_na_node_t node, const char*cmd,
					       padico_task_notifier_t notify, void*key)
  {
    padico_req_t req = padico_tm_req_new(notify, key);
    padico_control_send_ext(node, cmd, req->id);
  }

  /** Send a blocking request on the control channel. */
  static inline padico_req_t padico_control_send(padico_na_node_t node, const char*cmd)
  {
    padico_req_t req = padico_tm_req_new(NULL, NULL);
    padico_control_send_ext(node, cmd, req->id);
    return req;
  }

  /** Get the control channel as a network */
  static inline padico_topo_network_t padico_control_getnetwork(void)
  {
    static padico_topo_network_t net = NULL;
    if(net == NULL)
      net = padico_na_network_getbyid("ControlNetwork");
    return net;
  }

#endif /* __cplusplus */


  /* *** NetAccess Control *********************************** */
  
  struct padico_control_event_s
  {
    enum padico_control_event_kind_e
      {
	/** a new node appeared in the known topology */
	PADICO_CONTROL_EVENT_NEW_NODE         = 1,
	/** multiple new nodes appeared at once */
	PADICO_CONTROL_EVENT_NEW_NODE_VECT    = 2,
	/** a node was deleted/disconnected/became unreachable */
	PADICO_CONTROL_EVENT_DELETE_NODE      = 3,
	/** a new controler (PadicoControl) is loaded */
	PADICO_CONTROL_EVENT_PLUG_CONTROLER   = 4,
	/** a new connector (PadicoConnector) is loaded */
	PADICO_CONTROL_EVENT_PLUG_CONNECTOR   = 5,
	/** a new control selector (ControlSelector) is loaded */
	PADICO_CONTROL_EVENT_PLUG_SELECTOR    = 6,
	/** a controler (PadicoControl) is unloaded */
	PADICO_CONTROL_EVENT_UNPLUG_CONTROLER = 7,
	/** a connector (PadicoConnector) is unloaded */
	PADICO_CONTROL_EVENT_UNPLUG_CONNECTOR = 8,
	/** a control selector (ControlSelector) is unloaded */
	PADICO_CONTROL_EVENT_UNPLUG_SELECTOR  = 9,
	/** initial event sent to new controlers to get the past events and current topology state */
	PADICO_CONTROL_EVENT_INIT_CONTROLER   = 10,
	/** @internal force processing in the XML tasklet */
	PADICO_CONTROL_EVENT_PROCESSING       = 11
      } kind;
    union
    {
      struct
      {
	puk_instance_t controler;
	padico_na_node_vect_t nodes;
      } INIT_CONTROLER;
      struct
      {
	puk_instance_t controler;
	union
	{
	  padico_na_node_t node;
	  padico_na_node_vect_t nodes;
	};
      } NODE;
      struct
      {
	puk_instance_t controler;
	puk_instance_t server;    /**< the server instance that created the new controler; may be NULL */
      } PLUG_CONTROLER;
      struct
      {
	puk_adapter_t connector;
      } PLUG_CONNECTOR;
      struct
      {
	puk_adapter_t selector;
      } PLUG_SELECTOR;
      struct
      {
	puk_instance_t controler;
      } UNPLUG_CONTROLER;
      struct
      {
	puk_adapter_t connector;
      } UNPLUG_CONNECTOR;
      struct
      {
	puk_adapter_t selector;
      } UNPLUG_SELECTOR;
      struct /** Control command descriptor, in HOST byte order; describes a received request. */
      {
	padico_na_node_t node;         /**< sender of this request */
	char*cmd;                      /**< msg of the request */
	size_t size;                   /**< message size (including attachments) */
      } PROCESSING;
    };
  };
  typedef struct padico_control_event_s*padico_control_event_t;

  /* shortcuts for the event channel */

#ifndef __cplusplus

  static inline void padico_control_event_enable(void)
  {
    padico_tasklet_release(padico_control_get_event_channel());
  }
  static inline void padico_control_event_disable(void)
  {
    padico_tasklet_acquire(padico_control_get_event_channel());
  }
  static inline void padico_control_event_flush(void)
  {
    padico_tasklet_flush(padico_control_get_event_channel());
  }
  static inline void padico_control_event_sync(void)
  {
    padico_control_event_disable();
    padico_control_event_flush();
    padico_control_event_enable();
  }
  static inline void padico_control_event_subscribe(padico_event_listener_t l)
  {
    padico_control_event_disable();
    padico_control_gen_init_events(l);
    padico_tasklet_subscribe(padico_control_get_event_channel(), l);
    padico_control_event_flush();
    padico_control_event_enable();
  }
  static inline void padico_control_event_unsubscribe(padico_event_listener_t l)
  {
    padico_control_event_disable();
    padico_tasklet_unsubscribe(padico_control_get_event_channel(), l);
    padico_control_event_flush();
    padico_control_event_enable();
  }


  static inline void padico_control_new_node(puk_instance_t controler,
					     padico_na_node_t node)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_NEW_NODE;
    event->NODE.controler = controler;
    event->NODE.node      = node;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }
  static inline void padico_control_new_node_vect(puk_instance_t controler,
						  padico_na_node_vect_t nodes)
  {
    padico_na_node_vect_itor_t i;
    puk_vect_foreach(i, padico_na_node, nodes)
      {
	padico_control_new_node(controler, *i);
      }
  }
  static inline void padico_control_plug_controler(puk_instance_t controler, puk_instance_t server)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_PLUG_CONTROLER;
    event->PLUG_CONTROLER.controler = controler;
    event->PLUG_CONTROLER.server = server;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }

  static inline void padico_control_plug_connector(puk_adapter_t connector)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_PLUG_CONNECTOR;
    event->PLUG_CONNECTOR.connector = connector;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }

  static inline void padico_control_plug_selector(puk_adapter_t selector)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_PLUG_SELECTOR;
    event->PLUG_SELECTOR.selector = selector;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }
  static inline void padico_control_unplug_controler(puk_instance_t controler)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_UNPLUG_CONTROLER;
    event->UNPLUG_CONTROLER.controler = controler;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }
  static inline void padico_control_unplug_connector(puk_adapter_t connector)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_UNPLUG_CONNECTOR;
    event->UNPLUG_CONNECTOR.connector = connector;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }

  static inline void padico_control_unplug_selector(puk_adapter_t selector)
  {
    struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
    event->kind = PADICO_CONTROL_EVENT_UNPLUG_SELECTOR;
    event->UNPLUG_SELECTOR.selector = selector;
    padico_tasklet_schedule(padico_control_get_event_channel(), event);
  }

  static inline void padico_control_delete_node(puk_instance_t controler,
						padico_na_node_t node)
  {
    if(node != NULL)
      {
	struct padico_control_event_s*event = padico_malloc(sizeof(struct padico_control_event_s));
	event->kind = PADICO_CONTROL_EVENT_DELETE_NODE;
	event->NODE.controler = controler;
	event->NODE.node      = node;
	padico_tasklet_schedule(padico_control_get_event_channel(), event);
      }
  }

#endif /* __cplusplus */

  /* *** PadicoControl */

  /* Interface: "PadicoControl" */
  struct padico_na_control_driver_s
  {
    void (*send_common)(void*_status, padico_na_node_t node, padico_control_msg_t msg);
  };
  PUK_IFACE_TYPE(PadicoControl, struct padico_na_control_driver_s);

  /* Interface: "PadicoConnector" */
  struct padico_control_connector_driver_s
  {
    /** establishes a secondary control connection, to an already reachable node.
     * request 'req' is signaled upon completion.
     * Expects both peers do a 'connect' */
    int (*connect)(void*_status, padico_req_t req, padico_na_node_t node);
  };
  PUK_IFACE_TYPE(PadicoConnector, struct padico_control_connector_driver_s);

  /* Interface: "PadicoBootstrap" */

  struct padico_bootstrap_addr_s
  {
    struct in_addr ip;
    uint16_t port;
  };
  struct padico_bootstrap_driver_s
  {
    /** create a new bootstrap endpoint. 
     * 'ip' must be INADDR_ANY, no other value supported yet
     * 'port' may be a port number or 0 for ephemeral port, parameter is filled with actual port (in network byte order)
     */
    int (*bootstrap_new)    (void*_status, const struct in_addr*ip, uint16_t*port);
    /** connect to a listening bootstrap endpoint */
    int (*bootstrap_connect)(void*_status, const struct in_addr*ip, const uint16_t port);
  };
  PUK_IFACE_TYPE(PadicoBootstrap, struct padico_bootstrap_driver_s);

  /* Interface: "ControlSelector" */
  struct padico_control_selector_driver_s
  {
    /** returns an instance with a PadicoControl facet */
    puk_instance_t (*router)(padico_na_node_t node);
  };
  PUK_IFACE_TYPE(ControlSelector, struct padico_control_selector_driver_s);


  /** @} */


#include "na-Group.h"


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* PADICO_REMOTECONTROL_H */

