/*
 * NewMadeleine
 * Copyright (C) 2014 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* -*- Mode: C; tab-width: 2; c-basic-offset: 2 -*- */

/** @file Shared-memory driver with 'minidriver' interface.
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <nm_minidriver.h>
#include "shm.h"

/* ********************************************************* */

static void*nm_minidriver_shm_instanciate(puk_instance_t instance, puk_context_t context);
static void nm_minidriver_shm_destroy(void*);

static const struct puk_adapter_driver_s nm_minidriver_shm_adapter =
  {
    .instanciate = &nm_minidriver_shm_instanciate,
    .destroy = &nm_minidriver_shm_destroy
  };

static void nm_minidriver_shm_getprops(int index, struct nm_minidriver_properties_s*props);
static void nm_minidriver_shm_init(puk_context_t context, const void**drv_url, size_t*url_size);
static void nm_minidriver_shm_connect(void*_status, const void*remote_url, size_t url_size);
static void nm_minidriver_shm_send_post(void*_status, const struct iovec*v, int n);
static int  nm_minidriver_shm_send_poll(void*_status);
static void nm_minidriver_shm_recv_init(void*_status,  struct iovec*v, int n);
static int  nm_minidriver_shm_poll_one(void*_status);
static int  nm_minidriver_shm_cancel_recv(void*_status);

static const struct nm_minidriver_iface_s nm_minidriver_shm_minidriver =
  {
    .getprops    = &nm_minidriver_shm_getprops,
    .init        = &nm_minidriver_shm_init,
    .connect     = &nm_minidriver_shm_connect,
    .send_post   = &nm_minidriver_shm_send_post,
    .send_poll   = &nm_minidriver_shm_send_poll,
    .recv_init   = &nm_minidriver_shm_recv_init,
    .poll_one    = &nm_minidriver_shm_poll_one,
    .cancel_recv = &nm_minidriver_shm_cancel_recv
  };

/* ********************************************************* */

PADICO_MODULE_COMPONENT(Minidriver_shm,
  puk_component_declare("Minidriver_shm",
			puk_component_provides("PadicoAdapter", "adapter", &nm_minidriver_shm_adapter),
			puk_component_provides("NewMad_minidriver", "minidriver", &nm_minidriver_shm_minidriver)));

/* ********************************************************* */


/** 'shm' per-context data.
 */
struct nm_minidriver_shm_context_s
{
  struct padico_shm_s*shm;                 /**< shm segment */
  struct nm_minidriver_shm_s*conns[PADICO_SHM_NUMNODES]; /**< connections in the context, indexed by destination rank */
};

/** 'shm' per-instance status.
 */
struct nm_minidriver_shm_s
{
  struct padico_shm_node_s*dest;           /**< destination node */
  int dest_rank;                           /**< destination rank in shm directory */
  struct nm_minidriver_shm_context_s*shm_context;
  struct padico_shm_short_lfqueue_s queue; /**< per-connection receive queue */
  struct
  {
    struct iovec*v;
    int n;
  } recv;
};


/** dispatch blocks from whole context to separate connections */
static void nm_minidriver_shm_dispatch(struct nm_minidriver_shm_context_s*shm_context)
{
  int block = padico_shm_short_lfqueue_dequeue(&shm_context->shm->self->mailbox.short_queue);
  if(block != -1)
    {
      struct padico_shm_short_header_s*header = padico_shm_block_get_ptr(shm_context->shm, block);
      const int from = header->from;
      assert(from >= 0);
      assert(from < PADICO_SHM_NUMNODES);
      struct nm_minidriver_shm_s*conn = shm_context->conns[from];
      assert(conn != NULL);
      padico_shm_short_lfqueue_enqueue(&conn->queue, block);
    }
}

/* ********************************************************* */

static void*nm_minidriver_shm_instanciate(puk_instance_t instance, puk_context_t context)
{  
  struct nm_minidriver_shm_s*status = padico_malloc(sizeof(struct nm_minidriver_shm_s));
  status->dest = NULL;
  status->dest_rank = -1;
  padico_shm_short_lfqueue_init(&status->queue);
  status->shm_context = puk_context_get_status(context);
  assert(status->shm_context->shm->seg != NULL);
  return status;

}

static void nm_minidriver_shm_destroy(void*_status)
{
  struct nm_minidriver_shm_s*status = _status;
  if(status->dest_rank > -1)
    status->shm_context->conns[status->dest_rank] = NULL;
  padico_free(_status);
}

/* ********************************************************* */

static void nm_minidriver_shm_getprops(int index, struct nm_minidriver_properties_s*props)
{
  props->profile.latency = 200;
  props->profile.bandwidth = 10000;
}

static void nm_minidriver_shm_init(puk_context_t context, const void**drv_url, size_t*url_size)
{
  struct nm_minidriver_shm_context_s*shm_context = padico_malloc(sizeof(struct nm_minidriver_shm_context_s));
  padico_string_t segment_name = padico_string_new();
  padico_string_printf(segment_name, "minidriver-id%s", context->id);
  shm_context->shm = padico_shm_init(padico_string_get(segment_name));
  int i;
  for(i = 0; i < PADICO_SHM_NUMNODES; i++)
    {
      shm_context->conns[i] = NULL;
    }
  puk_context_set_status(context, shm_context);
  const char*url = padico_strdup((char*)padico_na_node_getuuid(padico_na_getlocalnode()));
  *drv_url = url;
  *url_size = strlen(url);
}

static void nm_minidriver_shm_connect(void*_status, const void*remote_url, size_t url_size)
{ 
  struct nm_minidriver_shm_s*status = (struct nm_minidriver_shm_s*)_status;
  const char*remote_uuid = remote_url;
  padico_na_node_t remote_node = padico_na_getnodebyuuid((padico_topo_uuid_t)remote_uuid);
  assert(status->shm_context->shm->seg != NULL);
  status->dest = padico_shm_directory_node_lookup(status->shm_context->shm, remote_node);
  if(status->dest == NULL)
    {
      padico_fatal("minidriver-shm: FATAL- cannot find peer node %s in shm directory.\n", remote_uuid);
    }
  else
    {
      status->dest_rank = status->dest->rank;
      assert(status->dest_rank >= 0);
      assert(status->dest_rank < PADICO_SHM_NUMNODES);
      status->shm_context->conns[status->dest_rank] = status;
    }
}

static void nm_minidriver_shm_send_post(void*_status, const struct iovec*v, int n)
{
  struct nm_minidriver_shm_s*status = _status;
  const int block_num = padico_shm_block_get(status->shm_context->shm);
  void*base = padico_shm_short_get_ptr(status->shm_context->shm, block_num);
  size_t len = 0;
  int i;
  for(i = 0; i < n; i++)
    {
      if(len + v[i].iov_len > PADICO_SHM_SHORT_PAYLOAD)
	padico_fatal("minidriver-shm: FATAL- cannot send, payload too large for short buffer.\n");
      memcpy(base + len, v[i].iov_base, v[i].iov_len);
      len += v[i].iov_len;
    }
  assert(len <= PADICO_SHM_SHORT_PAYLOAD);
  padico_shm_short_send_commit(status->shm_context->shm, status->dest, block_num, len);
}

static int nm_minidriver_shm_send_poll(void*_status)
{
  return NM_ESUCCESS;
}


static void nm_minidriver_shm_recv_init(void*_status, struct iovec*v, int n)
{
  struct nm_minidriver_shm_s*status = _status;
  status->recv.v = v;
  status->recv.n = n;
}

static int nm_minidriver_shm_poll_one(void*_status)
{
  struct nm_minidriver_shm_s*status = _status;
  int block = padico_shm_short_lfqueue_dequeue_single_reader(&status->queue);
  if(block == -1)
    {
      nm_minidriver_shm_dispatch(status->shm_context);
      block = padico_shm_short_lfqueue_dequeue(&status->queue);
    }
  if(block != -1)
    {
      struct padico_shm_short_header_s*header = padico_shm_block_get_ptr(status->shm_context->shm, block);
      void*buffer = header + 1;
      assert(status->recv.n == 1);
      if(status->recv.v->iov_len < header->size)
	{
	  return -NM_EINVAL;
	}
      else
	{
	  memcpy(status->recv.v->iov_base, buffer, header->size);
	  status->recv.v = NULL;
	  status->recv.n = -1;
	  padico_shm_block_free(status->shm_context->shm, block);
	  return NM_ESUCCESS;
	}
    }
  else
    {
      return -NM_EAGAIN;
    }
}

static int nm_minidriver_shm_cancel_recv(void*_status)
{ 
  return -NM_ENOTIMPL;
}
