/** @file
 */

#ifndef PADICO_SHM_H
#define PADICO_SHM_H

#include <sys/types.h>
#include <semaphore.h>
#include <unistd.h>

#define SHM_USE_POSIX

#if defined(SHM_USE_POSIX)
#  include <sys/mman.h>
#  include <fcntl.h>
#elif defined(SHM_USE_IPC)
#  warning "Using IPC shm"
#  include <sys/shm.h>
#else
#  error "No shm method defined"
#endif

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PSP.h>

#include "spinlock.h"


/* ********************************************************* */

#define PADICO_SHM_NUMNODES  128
/** Number of blocks allocated in shm buffer
 * @note to prevent deadlocks, should be at least num_nodes+1
 */
#define PADICO_SHM_NUMBLOCKS 4096
#define PADICO_SHM_BLOCKSIZE (16 * 1024)
#define PADICO_SHM_QUEUESIZE 128
#define PADICO_SHM_NUMQUEUES 128
#define PADICO_SHM_BLOCKCACHESIZE 32
#define PADICO_YIELD_ON_WAIT


/** type for short messages queues- contains block index */
PUK_LFQUEUE_TYPE(padico_shm_short, int, -1, PADICO_SHM_QUEUESIZE);
/** type for free blocks queue- contains block index */
PUK_LFQUEUE_TYPE(padico_shm_block, int, -1, PADICO_SHM_NUMBLOCKS*2);
/** type for free blocks cache- contains block index */
PUK_LFQUEUE_TYPE(padico_shm_blockcache, int, -1, PADICO_SHM_BLOCKCACHESIZE);
/** type for large message queues- contains block index */
PUK_LFQUEUE_TYPE(padico_shm_large, int, -1, PADICO_SHM_QUEUESIZE);
/** type for queues of queues- contains large queues index */
PUK_LFQUEUE_TYPE(padico_shm_queue, int, -1, PADICO_SHM_NUMQUEUES*2);

struct padico_shm_node_s
{
  struct
  {
    struct padico_shm_short_lfqueue_s short_queue; /**< queue for short messages sent eagerly; contains block index */
    struct padico_shm_blockcache_lfqueue_s block_cache; /**< free blocks cache */
    struct
    {
      volatile int pipeline;  /**< allocated queue for transfer */
    } large[PADICO_SHM_NUMNODES]; /**< large messages mbox; unlocked acces, valid when pipeline != -1 */
  } mailbox;
  char node_uuid[PADICO_TOPO_UUID_SIZE + 1];
  int pid;
  int rank;
  volatile int enabled;
};

/** The mmaped segment
 */
struct padico_shm_seg_s
{
  struct
  {
    char date[16];
    char time[16];
  } build_string;  /**< check that all nodes use the same build */
  struct
  {
    struct padico_shm_node_s nodes[PADICO_SHM_NUMNODES];
    volatile int num_nodes;
  } directory;
  struct
  {
    char block[PADICO_SHM_BLOCKSIZE * PADICO_SHM_NUMBLOCKS];
    struct padico_shm_block_lfqueue_s free_blocks;
    struct padico_shm_large_lfqueue_s queues[PADICO_SHM_NUMQUEUES]; /**< pool of queues to pipeline large transfers */
    struct padico_shm_queue_lfqueue_s free_queues;
    struct padico_shm_block_lfqueue_s pending_blocks[PADICO_SHM_NUMNODES];    /**< used only by nm_strat_shm */
  } buffer;
};

/** Generic shm header for short messages in the eager queue
 */
struct padico_shm_short_header_s
{
  volatile uint32_t from;  /**< index of sender in directory */
  volatile uint32_t size;  /**< actual size of data payload in the message (excluding header size) */
};

#define PADICO_SHM_SHORT_PAYLOAD (PADICO_SHM_BLOCKSIZE - sizeof(struct padico_shm_short_header_s))

/** global state of the Shm segment
 */
__PUK_SYM_INTERNAL
struct padico_shm_s
{
  struct padico_shm_seg_s*seg;
  struct padico_shm_node_s*self;
  int rank;
  char*shm_path;
  char*sem_path;
};

/* ********************************************************* */

extern struct padico_shm_s*padico_shm_init(const char*name);

extern void padico_shm_close(struct padico_shm_s*shm);

extern struct padico_shm_node_s*padico_shm_directory_node_lookup(struct padico_shm_s*shm, padico_na_node_t node);

extern void padico_shm_large_send(struct padico_shm_s*shm, const void*bytes, size_t size, struct padico_shm_node_s*dest);
extern int  padico_shm_large_recv_post(struct padico_shm_s*shm, int from);
extern int  padico_shm_large_recv_poll(struct padico_shm_s*shm, void*bytes, size_t size, int queue_num);
extern void padico_shm_large_recv(struct padico_shm_s*shm, void*bytes, size_t size, int from);


/* ********************************************************* */
/* ** shared memory blocks allocator */

/** allocate a memory block in shm buffer */
static inline int padico_shm_block_get(struct padico_shm_s*shm)
{
  int block = -1;
 retry:
  block = padico_shm_blockcache_lfqueue_dequeue_single_reader(&shm->self->mailbox.block_cache);
  if(block == -1)
    block = padico_shm_block_lfqueue_dequeue(&shm->seg->buffer.free_blocks);
  if(block == -1)
    {
      padico_warning("Shm: padico_shm_block_get()- out of space in shared memory segment; waiting...\n");
      usleep(1000);
      goto retry;
    }
  assert(block >= 0);
  assert(block < PADICO_SHM_NUMBLOCKS);
  return block;
}
/** get the pointer to a block given its block number */
static inline void*padico_shm_block_get_ptr(struct padico_shm_s*shm, int block_num)
{
  assert(block_num >= 0);
  assert(block_num < PADICO_SHM_NUMBLOCKS);
  return shm->seg->buffer.block + block_num * PADICO_SHM_BLOCKSIZE;
}
/** return a free block back to the free blocks pool */
static inline void padico_shm_block_free(struct padico_shm_s*shm, int block_num)
{
  assert(block_num >= 0);
  assert(block_num < PADICO_SHM_NUMBLOCKS);
  int rc = padico_shm_blockcache_lfqueue_enqueue_single_writer(&shm->self->mailbox.block_cache, block_num);
  if(rc)
    rc = padico_shm_block_lfqueue_enqueue(&shm->seg->buffer.free_blocks, block_num);
  if(rc)
    {
      padico_fatal("Shm: error while enqueuing free block.\n");
    }
}

/* ********************************************************* */
/* ** messaging */

static inline void padico_shm_short_send_commit(struct padico_shm_s*shm, struct padico_shm_node_s*dest, int block_num, size_t size)
{
  void*block = padico_shm_block_get_ptr(shm, block_num);
  struct padico_shm_short_header_s*shm_header = block;
  shm_header->from = shm->rank;
  shm_header->size = size;
  int rc = -1;
  while(rc)
    {
      rc = padico_shm_short_lfqueue_enqueue(&dest->mailbox.short_queue, block_num);
      if(rc)
	{
	  padico_warning("Shm: cannot send- receiver queue full.\n");
	  usleep(10);
	}
    }
}
static inline void*padico_shm_short_get_ptr(struct padico_shm_s*shm, int block_num)
{
  void*block = padico_shm_block_get_ptr(shm, block_num);
  return block + sizeof(struct padico_shm_short_header_s);
}
static inline int padico_shm_short_recv_poll(struct padico_shm_s*shm)
{
  int block_num = padico_shm_short_lfqueue_dequeue(&shm->self->mailbox.short_queue);
  return block_num;
}

/* ********************************************************* */

#endif /* PADICO_SHM_H */
