/** @file
 */

#ifndef SOCKETFACTORY_H
#define SOCKETFACTORY_H

#include <Padico/Topology.h>

typedef void (*padico_socketfactory_acceptor_t)(void*key, int fd, const struct sockaddr*addr, socklen_t addrlen);
typedef void (*padico_socketfactory_connector_t)(void*key, int fd);

struct padico_socketfactory_s
{
  /** create the socket attached to the SocketFactory instance
   * @return 0 if success, -errno in case of error.
   */
  int (*create) (void*_status, int family, int type, int protocol);
  /** binds the socket to the given address (expected: AF_INET family)
   * or performs autobind (AF_PADICO family + 'AUTO' magic number).
   * Socket factories are not required to support all address families.
   * @return 0 if success, -errno in case of error.
   * @note 'addr' and 'addrlen' are filled-in with the address.
   */
  int (*bind)   (void*_status, struct sockaddr*addr, socklen_t*addrlen);
  /** listens incoming connections on the socket
   * @return 0 if success, -errno in case of error.
   * @param addr address of the expected client (when used as secondary adapter).
   * @note when addr is non-NULL (==>secondary), 'backlog' is ignored.
   * @param notifier notification function called upon connection completion
   * @param key key given to the notifier
   */
  int (*listen) (void*_status, int backlog, const struct sockaddr*addr, socklen_t addrlen,
		 padico_socketfactory_acceptor_t notifier, void*key);
  /** connects the socket to the given address
   * @return 0 if success, -errno in case of error.
   * @param addr address of the server.
   * @param notifier notification function called upon connection completion
   * @param key key given to the notifier
   */
  int (*connect)(void*_status, const struct sockaddr*addr, socklen_t addrlen,
		 padico_socketfactory_connector_t notifier, void*key);
  /** closes the given socket
   */
  int (*close)(void*_status);
  /** sets socket options (if appropriate for adapter).
   */
  int (*setsockopt)(void*_status, int level, int opt, const void*val, int len);
  /** socket capabilities (@see enum padico_vsock_capabilities_e)
   */
  unsigned caps; 
};

PUK_IFACE_TYPE(SocketFactory, struct padico_socketfactory_s);

#endif /* SOCKETFACTORY_H */

