/** @file
 * @brief Puk hashtables
 * @ingroup Puk
 */

#include "Puk.h"
#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
#include <sys/types.h>
#endif
#include <assert.h>
#include <string.h>

#ifndef PUK_HASHTABLE_H
#define PUK_HASHTABLE_H

typedef uint32_t (*puk_hash_hfunc_t)(const void*key);
typedef int      (*puk_hash_eqfunc_t)(const void*key1, const void*key2);

/* ********************************************************* */
/* *** hashtable */

typedef struct puk_hashtable_s*puk_hashtable_t;

puk_hashtable_t puk_hashtable_new(puk_hash_hfunc_t hash_func,
				  puk_hash_eqfunc_t eq_func);
void puk_hashtable_delete(puk_hashtable_t h);
void puk_hashtable_insert(puk_hashtable_t h, const void*key, void*data);
void*puk_hashtable_lookup(puk_hashtable_t h, const void*key);
void puk_hashtable_lookup2(puk_hashtable_t h, const void*key, void*outkey, void*data);
int  puk_hashtable_probe(puk_hashtable_t h, const void*key);
void puk_hashtable_remove(puk_hashtable_t h, const void*key);
int  puk_hashtable_size  (puk_hashtable_t h);
void puk_hashtable_getentry(puk_hashtable_t h, void*key, void*data);

/* ********************************************************* */
/* *** hashtable enumerator */

struct puk_hashtable_enumerator_s
{
  puk_hashtable_t h;
  struct puk_hash_entry_s*current;
  int i;
};
typedef struct puk_hashtable_enumerator_s*puk_hashtable_enumerator_t;

void puk_hashtable_enumerator_init(puk_hashtable_enumerator_t e, puk_hashtable_t h);
void puk_hashtable_enumerator_next2(puk_hashtable_enumerator_t e, void*key, void*data);

static inline puk_hashtable_enumerator_t puk_hashtable_enumerator_new(puk_hashtable_t h)
{
  puk_hashtable_enumerator_t e = (struct puk_hashtable_enumerator_s*)
    padico_malloc(sizeof(struct puk_hashtable_enumerator_s));
  puk_hashtable_enumerator_init(e, h);
  return e;
}
static inline void puk_hashtable_enumerator_delete(puk_hashtable_enumerator_t e)
{
  padico_free(e);
}

static inline const void*puk_hashtable_enumerator_next_key(puk_hashtable_enumerator_t e)
{
  const void*_key;
  puk_hashtable_enumerator_next2(e, &_key, NULL);
  return _key;
}
static inline void*puk_hashtable_enumerator_next_data(puk_hashtable_enumerator_t e)
{
  void*_data;
  puk_hashtable_enumerator_next2(e, NULL, &_data);
  return _data;
}

/* ********************************************************* */
/* *** hashing functions */

uint32_t puk_hash_oneatatime(const unsigned char*key, int len);
int      puk_hash_string_eq(const void*key1, const void*key2);
uint32_t puk_hash_string(const void*key);
int      puk_hash_pointer_eq(const void*key1, const void*key2);
uint32_t puk_hash_pointer(const void*key);
int      puk_hash_int_eq(const void*key1, const void*key2);
uint32_t puk_hash_int(const void*key);

/** creates a new hashtable with keys as strings */
#define puk_hashtable_new_string() \
  puk_hashtable_new(&puk_hash_string, &puk_hash_string_eq)

/** creates a new hashtable with keys as pointers */
#define puk_hashtable_new_ptr() \
  puk_hashtable_new(&puk_hash_pointer, &puk_hash_pointer_eq)

/** creates a new hashtable with keys as integers */
#define puk_hashtable_new_int() \
  puk_hashtable_new(&puk_hash_int, &puk_hash_int_eq)

#endif /* PUK_HASTABLE_H */



