/** @file
 * @brief Provides "binary" module driver
 * @ingroup Puk
 */

#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);


/* ********************************************************* */

static padico_rc_t unit_binary_load  (puk_unit_t unit);
static padico_rc_t unit_binary_start (puk_unit_t unit, puk_job_t job);
static padico_rc_t unit_binary_unload(puk_unit_t unit);

const static struct padico_loader_s binary_driver =
{
  .name   = "binary",
  .load   = &unit_binary_load,
  .start  = &unit_binary_start,
  .unload = &unit_binary_unload
};


/* ********************************************************* */

__PUK_SYM_INTERNAL
void padico_puk_binary_init(void)
{
  puk_component_declare("PadicoLoader-binary", puk_component_provides("PadicoLoader", "loader", &binary_driver));
}

static padico_rc_t unit_binary_load(puk_unit_t unit)
{ 
  padico_rc_t rc = NULL;
  puk_lib_t lib = NULL;
  if(unit->driver_specific)
    {
      rc = padico_rc_error("binary driver: load error- unit **%s** is already loaded",
			   unit->name);
      return rc;
    }
  rc = puk_lib_open(&lib, unit->name);
  if (padico_rc_iserror(rc))
    {
      padico_warning("PadicoTM: *%s* load error.\n", unit->name);
      padico_rc_show(rc);
    }
  else
    {
      unit->driver_specific = (void*)lib;
      puk_lib_setself(lib, unit->mod);
      int err = puk_lib_init(lib);
      if(err)
	{
	  padico_warning("PadicoTM: *%s* init error.\n", unit->name);
	  rc = padico_rc_error("binary driver: unit **%s** load error- module initialization rc=%d", 
			       unit->name, err);
	}
    }
  return rc;
}

static padico_rc_t unit_binary_start(puk_unit_t unit, puk_job_t job)
{
  padico_rc_t rc = NULL; /* rc for the *start* operation (not for the job) */
  puk_lib_t lib  = (puk_lib_t)unit->driver_specific;
  if(lib != NULL)
    {
      puk_run_fun_t fun  = puk_lib_run_resolve(lib);
      if(fun != NULL)
	{
	  int i;
	  padico_out(50, "starting unit %s\n", unit->name);
	  
	  if(!job->argv)
	    {
	      job->argc = 1;
	      job->argv = padico_malloc(2 * sizeof(char*));
	      job->argv[0] = padico_strdup(unit->name);
	      job->argv[1] = NULL;
	    }
	  
	  padico_out(50, "argc = %d\n", job->argc);
	  for(i = 0; i < job->argc + 1; i++)
	    padico_out(50, " %d : %s\n", i, job->argv[i]);
	  
	  int rc = (*(fun))(job->argc, job->argv);
	  if(rc)
	    {
	      job->rc = padico_rc_new();
	      job->rc->rc  = rc;
	      job->rc->msg = padico_string_new();
	    }
	  else
	    {
	      job->rc = padico_rc_ok();
	    }
	  puk_job_notify(job);
	}
      else
	{
	  rc = padico_rc_error("binary driver: cannot start unit **%s** (not runnable)",
			       unit->name);
	}
    }
  else
    {
      rc = padico_rc_error("binary driver: cannot start unit **%s** (not properly initialized)",
			   unit->name);
    }
  return rc;
}

static padico_rc_t unit_binary_unload(puk_unit_t unit)
{
  const puk_lib_t lib = (puk_lib_t)unit->driver_specific;
  padico_rc_t rc = puk_lib_unload(lib);
  unit->driver_specific = NULL;
  return rc;
}


