/** @file
 * @brief Low-level shared libraries management. 
 * Used for binary and boot drivers
 * @ingroup Puk
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "Puk-internals.h"
#include "Module.h"


PADICO_MODULE_HOOK(Puk);

static char*lib_path = NULL;

void padico_puk_lib_init()
{
  asprintf(&lib_path, "%s/var/mod/binary/", padico_root());
}

static inline void puk_dylib_lock(void)
{
  puk_lock_acquire();
}
static inline void puk_dylib_unlock(void)
{
  puk_lock_release();
}

/* ***** Puk dylib ****************************************** */

void*puk_dylib_open(const char *pathname, int mode)
{
  puk_dylib_lock();
  void*p = dlopen(pathname, mode);
  if(p == NULL)
    {  
      int retry = 10;
      while(p == NULL && retry-- > 0)
	{
	  usleep(100000);
	  p = dlopen(pathname, mode);
	}
    }
  puk_dylib_unlock();
  return p;
}

int puk_dylib_close(void *handle)
{
  puk_dylib_lock();
  int rc = dlclose(handle);
  puk_dylib_unlock();
  return rc;
}

char*puk_dylib_error(void)
{
  puk_dylib_lock();
  char*rc = dlerror();
  puk_dylib_unlock();
  return rc;
}

void*puk_dylib_resolve(void*handle, const char*symbol) 
{
  void*rc;
  const int symbol_maxlen = 256;
  char mangled_symbol[symbol_maxlen];
  snprintf(mangled_symbol, symbol_maxlen, "%s%s", SYMBOL_PREFIX, symbol);
  puk_dylib_lock();
  rc = dlsym(handle, mangled_symbol);
  if(rc == NULL)
    {
      rc = dlsym(handle, symbol);
    }
  puk_dylib_unlock();
  return rc;
}


/* ********************************************************** */

padico_rc_t puk_lib_open(puk_lib_t* lib, const char*lib_name)
{
  padico_rc_t rc = NULL;

  *lib = puk_lib_new();
  (*lib)->name = padico_strdup(lib_name);
  (*lib)->filename = padico_malloc(strlen(lib_path) + strlen(lib_name) + 2 + sizeof(long));
  strcpy((*lib)->filename, lib_path);
  strcat((*lib)->filename, lib_name);
  padico_out(20, "LIB LOAD loading %s\n", lib_name);
  
  if (access((*lib)->filename, R_OK) != 0)
    {
      rc = padico_rc_error("LIB load error: %s not found (%s file not found)", 
			   lib_name, (*lib)->filename);
      return rc;
    }
  
  (*lib)->dl_handle = puk_dylib_open((*lib)->filename, PUK_DYLIB_DEFAULT_MODE);
  if (!(*lib)->dl_handle) 
    {
      rc = padico_rc_error("LIB load error: puk_dylib_open() failed for %s (%s)", 
			   lib_name, puk_dylib_error());
      return rc;
    }
  padico_out(20, "LIB LOAD %s ok\n", lib_name);
  return rc;
}

void puk_lib_setself(puk_lib_t lib, puk_mod_t mod)
{
  typedef void (*setself_fun_t)(puk_mod_t);
  setself_fun_t setself;
  padico_out(40, "LIB INIT **%s**\n", lib->name);
  setself = (setself_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_setself");
  if (setself != NULL)
    {
      (*setself)(mod);
    }
}

int puk_lib_init(puk_lib_t lib)
{
  puk_init_fun_t init_fun;
  int rc = 0;
  init_fun = (puk_init_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_init");
  if (init_fun != NULL)
    {
      rc = (*init_fun)();
    }
  return rc;
}

puk_run_fun_t puk_lib_run_resolve(puk_lib_t lib)
{
  puk_run_fun_t fun_ptr;

  padico_out(50, "LIB RUN try resolve in %s\n", lib->name);
  if(lib == NULL)
    {
      return NULL;
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_run")) != NULL)
    {
      padico_out(40, "LIB RUN padico_module_run() found --PadicoTM binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "marcel_main")) != NULL)
    {
      padico_out(40, "LIB RUN marcel_main() found --Marcel binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "main")) != NULL)
    {
      padico_out(40, "LIB RUN main() found --standard binary code.\n");
    }
  else if( (fun_ptr = (puk_run_fun_t) puk_dylib_resolve(lib->dl_handle, "MAIN__")) != NULL)
    {
      padico_out(40, "LIB RUN MAIN__() found --standard FORTRAN code.\n");
    }
  else
    {
      padico_out(6, "no runnable function found in lib %s.\n", lib->name);
    }
  return fun_ptr;
}

padico_rc_t puk_lib_unload(puk_lib_t lib)
{
  puk_finalize_fun_t fun_ptr;
  padico_rc_t rc = NULL;
  int err = -1;
  
  if(lib == NULL)
    {
      rc = padico_rc_error("LIB UNLOAD: lib %s was not succesfully initialized."
			   " Cannot unload.\n", lib->name);
      return rc;
    }
  fun_ptr = (puk_finalize_fun_t) puk_dylib_resolve(lib->dl_handle, "padico_module_finalize");
  if (fun_ptr != NULL)
    {
      (*fun_ptr)();
    }
  padico_out(50, "LIB closing lib.\n");
  err = puk_dylib_close(lib->dl_handle);
  padico_out(50, "LIB UNLOAD %s (dlclose rc=%d)\n", lib->name, err);
  if(err)
    {
      char*msg = puk_dylib_error();
      padico_warning("puk_dylib_close() error *%s*\n", msg);
      rc = padico_rc_error("LIB unload error: dlclose() failed for lib %s", lib->name);
      return rc;
    }
  padico_out(50, "LIB UNLOAD %s ok\n",
	       lib->name);
  puk_lib_delete(lib);
  return rc;
}


/***************************************************/

void puk_lib_ctor(puk_lib_t l)
{
  l->name = NULL;
  l->filename = NULL;
  l->dl_handle = NULL;
}

void puk_lib_dtor(puk_lib_t l) 
{
  if(l->filename)
    padico_free(l->filename);
  if(l->name)
    padico_free(l->name);
}


