/** @file
 * @brief Low-level module management
 * @note This low-level implementation is not thread-safe
 * @ingroup Puk
 */


#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);

static void defmod_start_handler  (puk_parse_entity_t);
static void defmod_end_handler    (puk_parse_entity_t);
static void requires_end_handler  (puk_parse_entity_t);
static void unit_end_handler      (puk_parse_entity_t);
static void attr_end_handler      (puk_parse_entity_t);

/******************************************************/

static struct
{
  puk_hashtable_t          modl;   /**< modules hashtable- key: module name (string); data: puk_mod_t */
  padico_modID_vect_t mod_names;   /**< vector of module names- updated only on request */
  const char*          mod_path;   /**< path to directory where modules are stored */
} puk_mod = { .modl = NULL, .mod_path = NULL };


/******************************************************/



/* *** MOD ************************************************* */
/* actions:
 *   getbyname
 *   open
 *   resolve
 *   load
 *   run
 *   unload
 */

padico_modID_t puk_mod_getname(puk_mod_t mod)
{
  return mod->mod_name;
}

void puk_mod_set_ancestor(puk_mod_t mod, puk_mod_t ancestor)
{
  if(ancestor != NULL)
    {
      mod->inherits = ancestor;
      ancestor->used_by++;
    }
}

padico_modID_vect_t puk_mod_getmodIDs(void)
{
  if(puk_mod.modl)
    {
      puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(puk_mod.modl);
      const char*name;
      if(puk_mod.mod_names)
	padico_modID_vect_delete(puk_mod.mod_names);
      puk_mod.mod_names = padico_modID_vect_new();
      name = puk_hashtable_enumerator_next_key(e);
      while(name)
	{
	  padico_modID_vect_push_back(puk_mod.mod_names, name);
	  name = puk_hashtable_enumerator_next_key(e);
	}
      puk_hashtable_enumerator_delete(e);
    }
  return puk_mod.mod_names;
}

puk_mod_t puk_mod_getbyname(const char*mod_name)
{
  puk_mod_t m = NULL;
  puk_lock_acquire();
  if(puk_mod.modl != NULL)
    m = puk_hashtable_lookup(puk_mod.modl, mod_name);
  puk_lock_release();
  return m;
}

/* private function
 */
padico_rc_t puk_mod_open(puk_mod_t*mod, padico_string_t filename)
{
  struct puk_parse_entity_s e = puk_xml_parse_file(padico_string_get(filename), PUK_TRUST_LOCAL);
  /* TODO: hard-coded trust_level=PUK_TRUST_LOCAL ?!? */
  padico_rc_t rc = puk_parse_get_rc(&e);
  if(padico_rc_iserror(rc))
    {
      return rc;
    }
  if(!puk_parse_is(&e, "defmod"))
    {
      rc = padico_rc_error("Parse error: root entity is not \"defmod\" while opening module %s",
			   padico_string_get(filename));
    }
  else
    {
      *mod = puk_parse_get_content(&e);
    }
  return rc;

}

/** padico_puk_mod_open(mod, mod_name)
 *  open a module, loads its description from disk to memory.
 *  mod     : (out) the loaded mod
 *  mod_name: (in)  module name
 */
padico_rc_t padico_puk_mod_open(puk_mod_t*mod, const char*mod_name)
{
  padico_string_t filename = padico_string_new();
  padico_string_printf(filename, "%s/meta/%s.xml", puk_mod.mod_path, mod_name);
  padico_out(50, "module=*%s* path=*%s*\n", mod_name, padico_string_get(filename));
  padico_rc_t rc = puk_mod_open(mod, filename);
  padico_out(50, "module=*%s* done.\n", mod_name);
  padico_string_delete(filename);
  return rc;
}

/** padico_puk_mod_open_file(mod, file_name)
 *  open a module, loads its description from disk to memory.
 *  mod     : (out) the loaded mod
 *  mod_name: (in)  module file description
 */
padico_rc_t padico_puk_mod_open_file(puk_mod_t*mod, const char*file_name)
{
  padico_string_t full_path = padico_string_new();
  padico_string_cat(full_path, file_name);
  padico_trace("Puk: MOD OPEN file=**%s**\n", file_name);
  return puk_mod_open(mod, full_path);
}

/** Resolve a module by name from:
 *    - already loaded modules
 *    - builtin modules
 *    - modules on disk
 */
padico_rc_t padico_puk_mod_resolve(puk_mod_t*_mod, const char*mod_name)
{
  puk_mod_t mod = NULL;
  padico_rc_t rc = NULL;
  /* 1- already loaded */
  mod = puk_mod_getbyname(mod_name);
  /* 2- try built-in */
  if(mod == NULL)
    {
      puk_mod_builtin_tryload(&mod, mod_name);
    }
  /* 3- open from file */
  if(mod == NULL)
    {
      rc = padico_puk_mod_open(&mod, mod_name);
    }
  if(_mod)
    *_mod = mod;
  return rc;
}

/** padico_puk_mod_load(mod)
 *  mod: (in) mod descriptor for an *already opened* module
 */
padico_rc_t padico_puk_mod_load(puk_mod_t mod)
{
  padico_rc_t rc = NULL;

  puk_lock_acquire();
  mod->ref_count++;
  padico_out(40, "module=*%s*; ref_count=%d\n", mod->mod_name, mod->ref_count);
  if(mod->ref_count > 1)
    {
      puk_lock_release();
      return rc;
    }
  /* resolve all module references */
  while(padico_modID_vect_size(&mod->unresolved_deps) > 0)
    {
      char*id = (char*)padico_modID_vect_pop_back(&mod->unresolved_deps);
      padico_trace("resolving required mod **%s** for **%s**\n", id, mod->mod_name);
      puk_mod_t reqmod = puk_mod_getbyname(id);
      if(reqmod == NULL)
	{
	  padico_rc_t urc = padico_puk_mod_resolve(&reqmod, id);
	  if(!padico_rc_iserror(urc))
	    {
	      puk_mod_set_ancestor(reqmod, padico_module_self());
	      padico_trace("loading required mod **%s** for **%s**\n", reqmod->mod_name, mod->mod_name);
	      rc = padico_puk_mod_load(reqmod);
	    }
	  else
	    {
	      urc = padico_rc_error("MOD load error: required module **%s** not found", id);
	      rc = padico_rc_cat(rc, urc);
	      padico_rc_delete(urc);
	      puk_lock_release();
	      return rc;
	    }
	}
      puk_mod_vect_push_back(&mod->requires, reqmod);
      padico_free(id);
    }
  /* add reverse references to required modules */
  puk_mod_vect_itor_t r;
  puk_vect_foreach(r, puk_mod, &mod->requires)
    {
      (*r)->used_by++;
    }
  /* add reverse reference to module containing the driver */
  assert(mod->driver->home != NULL);
  mod->driver->home->used_by++;
  puk_lock_release();

  /* load units */
  puk_unit_vect_itor_t u;
  puk_vect_foreach(u, puk_unit, &mod->units)
    {
      padico_rc_t urc = NULL;
      const struct padico_loader_s*driver = puk_component_get_driver_PadicoLoader(mod->driver, NULL);
      if(driver->load)
	urc = (driver->load)(u);
      if(padico_rc_iserror(urc))
      {
	rc = padico_rc_cat(rc, urc);
	padico_rc_delete(urc);
      }
    }
  return rc;
}

/** padico_puk_mod_start(mod, argc, argv)
 */
padico_rc_t padico_puk_mod_start(puk_mod_t mod, puk_job_t job)
{
  puk_unit_vect_itor_t u;
  padico_rc_t rc = NULL;
  padico_out(40, "module=**%s**\n", mod->mod_name);

  const char*setcwd = puk_mod_local_getattr(mod, "PADICO_SETCWD");
  if(setcwd)
    {
      int err = chdir(setcwd);
      if(err)
	{
	  rc = padico_rc_error("Puk: cannot chdir to %s.", setcwd);
	  return rc;
	}
      else
	{
	  padico_print("Puk: set working directory: %s\n", setcwd);
	}
    }
  puk_vect_foreach(u, puk_unit, &mod->units)
    {
      const struct padico_loader_s*driver = puk_component_get_driver_PadicoLoader(mod->driver, NULL);
      if(driver->start)
	{
	  job->rc = NULL;
	  rc = (driver->start)(u, job);
	  if(padico_rc_iserror(rc))
	    {
	      /* error => unit not runnable => continue */
	      padico_rc_delete(rc);
	      rc = NULL;
	    }
	  else 
	    {
	      /* only *one* runnable unit allowed per module */
	      return rc;
	    }
	}
    }
  rc = padico_rc_error("Puk: no runnable unit found in module %s.", mod->mod_name);
  puk_job_notify(job);
  return rc;
}

void puk_job_notify(puk_job_t job)
{
  if(job->notify)
    {
      (*job->notify)(job->notify_key);
    }
}

/** padico_puk_mod_unload(mod)
 */
padico_rc_t padico_puk_mod_unload(puk_mod_t mod)
{
  padico_rc_t rc = NULL;

  padico_out(40, "MOD UNLOAD %s (now ref=%d)\n",  mod->mod_name, mod->ref_count);
  if(mod->used_by > 0)
    {
      padico_warning("Puk: trying to unload module *%s* still in use.\n", mod->mod_name);
      rc = padico_rc_error("module in use (used_by=%d)", mod->used_by);
      return rc;
    }
  mod->ref_count--;
  if(mod->ref_count > 0)
    {
      rc = padico_rc_error("module in use (ref_count=%d)", mod->ref_count);
      return rc;
    }
  if(mod->ref_count < 0)
    {
      rc = padico_rc_error("module already unloaded (ref_count=%d)", mod->ref_count);
      return rc;
    }
  /* TODO- destroy all components of this mod 
   * i.e. components with c->home == mod
   */
  puk_unit_vect_itor_t u;
  puk_vect_foreach(u, puk_unit, &mod->units)
    {
      padico_rc_t urc = NULL;
      const struct padico_loader_s*driver = puk_component_get_driver_PadicoLoader(mod->driver, NULL);
      if(driver->unload)
	urc = (driver->unload)(u);
      if(padico_rc_iserror(urc))
	{
	  rc = padico_rc_cat(rc, urc);
	  padico_rc_delete(urc);
	}
    }
  if(mod->own_instances > 0)
    {
      padico_warning("Puk: module %s has %d instances after finalizing.\n", mod->mod_name, mod->own_instances);
    }
  puk_mod_vect_itor_t r;
  puk_vect_foreach(r, puk_mod, &mod->requires)
    {
      (*r)->used_by--;
    }
  mod->driver->home->used_by--;
  if(mod->inherits)
    mod->inherits->used_by--;
  padico_trace("Puk: MOD UNLOAD %s unloaded\n", mod->mod_name);
  puk_mod_destroy(mod);
  return rc;
}

/* *** ATTR ************************************************ */
/* attributes management
 *   getattr
 *   local_getattr
 *   setattr
 */

const char*puk_mod_local_getattr(puk_mod_t mod, const char*label)
{
  puk_attr_t attr = puk_hashtable_lookup(mod->attrs, label);
  if(attr)
    return attr->value;
  return NULL;
}

const char*puk_mod_getattr(puk_mod_t mod, const char*label)
{
  if(!mod)
    mod = padico_module_self();
  const char*attr = puk_mod_local_getattr(mod, label);
  const char*value = (attr || !mod->inherits) ? attr : puk_mod_getattr(mod->inherits, label);
  if((mod == padico_module_self()) && !value)
    {
      value = padico_getenv(label);
    }
  return value;
}

void puk_mod_setattr(puk_mod_t mod, const char*label, const char*value)
{
  puk_attr_t attr = puk_hashtable_lookup(mod->attrs, label);
  if(attr)
    {
      if(value)
	{
	  padico_out(40, "replacing attr label = \"%s\"\n", label);
	  padico_free((char*)attr->value);
	  attr->value = padico_strdup(value);
	}
      else
	{
	  puk_hashtable_remove(mod->attrs, label);
	}
    }
  else if(value)
    {
      padico_out(40, "storing attr label = \"%s\"\n", label);
      attr = padico_malloc(sizeof(struct puk_attr_s));
      attr->label = padico_strdup(label);
      attr->value = padico_strdup(value);
      puk_hashtable_insert(mod->attrs, attr->label, attr);
    }
}

void puk_mod_attr2argv(puk_mod_t mod, int *pargc, const char*const**pargv)
{
  /* find the number of argument */
  int i;
  int argc=1;
  char argname[]="ARG??";
  for(argc = 1; argc < 100; argc++)
    {
      const char *v;
      sprintf(&argname[3], "%d", argc);
      padico_trace("Puk: attr2argv arg %d: %s\n",argc, argname);
      v = puk_mod_local_getattr(mod, argname);
      if (v == NULL)
	break;
    }
  if(argc == 100)
    { 
      padico_fatal("Module %s has too many arguments (>99) !\n", mod->mod_name);
    }
  padico_trace("Puk: attr2argv found #arg %d\n",argc);

  const char**argv = padico_malloc((argc+1)*sizeof(char*));
  /* fill argv */
  argv[0] = padico_strdup(mod->mod_name);
  for(i = 1; i < argc; i++)
    {
      sprintf(&argname[3],"%d",i);
      argv[i] = puk_mod_local_getattr(mod, argname);
      padico_trace("Puk: attr2argv argv %d: %s\n", i, argv[i]);
    }
  argv[argc] = NULL;
  /* output result */
  *pargc = argc;
  *pargv = argv;
}

/***************************************************/

/* private func
 * allocates a new "puk_mod_t" object with default values
 */
puk_mod_t puk_mod_new(const char*name, const char*driver_name)
{
  puk_mod_t m = padico_malloc(sizeof(struct puk_mod_s));
  /* resolve driver */
  padico_string_t drv_mod_name = padico_string_new();
  padico_string_printf(drv_mod_name, "PadicoLoader-%s", driver_name);
  puk_component_t driver = puk_adapter_resolve(padico_string_get(drv_mod_name));
  padico_string_delete(drv_mod_name);
  if(!driver)
    {
      padico_warning("Puk: driver **%s** not found for module **%s**\n", driver_name, name);
      return NULL; 
    }
  m->driver      = driver;
  m->mod_name    = padico_strdup(name);
  m->ref_count   = 0;
  m->used_by     = 0;
  m->managed_instances = 0;
  m->own_instances = 0;
  puk_mod_vect_init(&m->requires);
  padico_modID_vect_init(&m->unresolved_deps);
  m->inherits    = NULL;
  puk_unit_vect_init(&m->units);
  m->attrs       = puk_hashtable_new_string();
  puk_lock_acquire();
  if(puk_mod.modl == NULL)
    puk_mod.modl = puk_hashtable_new_string();
  puk_hashtable_insert(puk_mod.modl, m->mod_name, m);
  puk_lock_release();
  return m;
}

void puk_mod_destroy(puk_mod_t m) 
{
  assert(m->used_by == 0);
  assert(m->ref_count == 0);
  assert(m->managed_instances == 0);
  puk_lock_acquire();
  puk_hashtable_remove(puk_mod.modl, m->mod_name);
  puk_lock_release();
  if(m->attrs)
    {
      puk_hashtable_delete(m->attrs); /* TODO: free table content. */
      m->attrs = NULL;
    }
  puk_unit_vect_itor_t u;
  puk_vect_foreach(u, puk_unit, &m->units)
    {
      if(u->driver_specific)
	{
	  padico_warning("Puk: driver specific data not de-allocated for unit %s (driver: %s).",
			 u->name, m->driver->name);
	}
      if(u->name)
	{
	  padico_free(u->name);
	  u->name = NULL;
	}
    }
  puk_unit_vect_destroy(&m->units);
  puk_mod_vect_destroy(&m->requires);
  padico_modID_vect_destroy(&m->unresolved_deps);
  padico_free(m->mod_name);
  m->mod_name = NULL;
  padico_free(m);
}



/* *** parser funcs: defmod ******************************** */

static void defmod_start_handler(puk_parse_entity_t e)
{
  const char*mod_name = puk_parse_getattr(e, "name");
  const char*driver_name = puk_parse_getattr(e, "driver");
  puk_mod_t mod = NULL;
  if(!mod_name)
    {
      puk_parse_set_rc(e, padico_rc_error("no module name in <defmod/>\n"));
      return;
    }
  if(!driver_name)
    {
      puk_parse_set_rc(e, padico_rc_error("no driver name in <defmod/> for module %s.\n",
					  mod_name));
      return;
    }
  mod = puk_mod_new(mod_name, driver_name);
  puk_parse_set_content(e, mod);
}
static void defmod_end_handler(puk_parse_entity_t e)
{
  if(puk_parse_get_content(e) == NULL)
    {
      puk_parse_set_rc(e, padico_rc_error("error while loading module"));
    }
}
const static struct puk_tag_action_s defmod_action =
{
  .xml_tag        = "defmod",
  .start_handler  = &defmod_start_handler,
  .end_handler    = &defmod_end_handler,
  .required_level = PUK_TRUST_CONTROL
};

/* *** parser funcs: requires ****************************** */

static void requires_end_handler(puk_parse_entity_t e)
{ 
  puk_parse_entity_t owner = puk_parse_parent(e);
  puk_mod_t mod = NULL;
  padico_modID_t req = padico_strdup(puk_parse_get_text(e));
  padico_out(50, "requires=%s\n", req);
  if(owner == NULL)
    {
      padico_warning("Puk: ignoring <requires> tag\n");
      return;
    }
  mod = puk_parse_get_content(owner);
  padico_modID_vect_push_back(&mod->unresolved_deps, req);
}
const static struct puk_tag_action_s requires_action =
{
  .xml_tag        = "requires",
  .start_handler  = NULL,
  .end_handler    = &requires_end_handler,
  .required_level = PUK_TRUST_OTHER
};

/* *** parser funcs: unit ********************************** */

static void unit_end_handler(puk_parse_entity_t e)
{
  puk_parse_entity_t owner = puk_parse_parent(e);
  puk_mod_t mod = owner?puk_parse_get_content(owner):NULL;
  struct puk_unit_s unit = 
    {
      .name            = padico_strdup(puk_parse_get_text(e)),
      .mod             = mod,
      .driver_specific = NULL
    };
  padico_out(50, "unit=%s\n", unit.name);
  if(owner == NULL)
    {
      padico_warning("Puk: ignoring <unit> tag\n");
      return;
    }
  puk_unit_vect_push_back(&mod->units, unit);
}
const static struct puk_tag_action_s unit_action =
{
  .xml_tag        = "unit",
  .start_handler  = NULL,
  .end_handler    = &unit_end_handler,
  .required_level = PUK_TRUST_OTHER
};

/* *** parser funcs: file ********************************** */

const static struct puk_tag_action_s file_action =
{
  .xml_tag        = "file",
  .start_handler  = NULL,
  .end_handler    = NULL,
  .required_level = PUK_TRUST_OTHER
};

/* *** parser funcs: attr ********************************** */

static void attr_end_handler(puk_parse_entity_t e)
{
  const char*label = puk_parse_getattr(e, "label");
  if(label)
    {
      puk_parse_entity_t owner = puk_parse_parent(e);
      if(!owner)
	{
	  puk_parse_set_rc(e, padico_rc_error("attribute in wrong context"));
	}
      else
	{
	  const char*value = puk_parse_get_text(e);
	  puk_mod_t mod = puk_parse_get_content(owner);
	  puk_mod_setattr(mod, label, value);
	}
    }
  else
    {
      puk_parse_set_rc(e, padico_rc_error("<attr/> has no label.\n"));
      padico_warning("Puk: parsing error- <attr/> has no label.\n");
    }

}
const static struct puk_tag_action_s attr_action =
{
  .xml_tag        = "attr",
  .start_handler  = NULL,
  .end_handler    = &attr_end_handler,
  .required_level = PUK_TRUST_OTHER
};

/* *** parser funcs: setattr ******************************* */

static void setattr_end_handler(puk_parse_entity_t e)
{
  const char*label = puk_parse_getattr(e, "label");
  const char*mod_name = puk_parse_getattr(e, "mod")?:"Puk";
  if(label)
    {
      puk_mod_t mod = puk_mod_getbyname(mod_name);
      if(!mod)
	{
	  puk_parse_set_rc(e, padico_rc_error("module %s not found", mod_name));
	}
      else
	{
	  const char*value = puk_parse_get_text(e);
	  puk_mod_setattr(mod, label, value);
	}
    }
  else
    {
      puk_parse_set_rc(e, padico_rc_error("<setattr/> requires 'label' attribute.\n"));
      padico_warning("Puk: parsing error- <setattr/> failed.\n");
    }

}
const static struct puk_tag_action_s setattr_action =
{
  .xml_tag        = "setattr",
  .start_handler  = NULL,
  .end_handler    = &setattr_end_handler,
  .required_level = PUK_TRUST_OTHER
};

/******************************************************/

/**
 * @pre padico_puk_findcore_init() already done
 * @pre padico_puk_xml_init() already done
 */
void padico_puk_mod_init(void)
{
  /* init module database */
  if(puk_mod.modl == NULL)
    puk_mod.modl = puk_hashtable_new_string();
  puk_mod.mod_names = NULL;
  /* compute module path */
  padico_string_t s_path = padico_string_new();
  padico_string_catf(s_path, "%s/var/mod/", padico_root());
  puk_mod.mod_path = padico_strdup(padico_string_get(s_path));
  padico_string_delete(s_path);
  /* register parsing actions for module definition */
  puk_xml_add_action(defmod_action);
  puk_xml_add_action(requires_action);
  puk_xml_add_action(unit_action);
  puk_xml_add_action(file_action);
  puk_xml_add_action(attr_action);
  puk_xml_add_action(setattr_action);
}

void padico_puk_mod_finalize(void)
{
  return;
  puk_lock_acquire();
  padico_free((char*)puk_mod.mod_path);
  puk_mod.mod_path = NULL;
  while(puk_hashtable_size(puk_mod.modl) > 0)
    {
      puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(puk_mod.modl);
      puk_mod_t m = puk_hashtable_enumerator_next_data(e);
      while(m != NULL && (m->used_by > 0 || m->managed_instances > 0))
	m = puk_hashtable_enumerator_next_data(e);
      if(m == NULL)
	{
	  padico_print("Puk: dependency inconsistency detected while finalizing.\n");
	  e = puk_hashtable_enumerator_new(puk_mod.modl);
	  m = puk_hashtable_enumerator_next_data(e);
	  while(m != NULL)
	    {
	      m = puk_hashtable_enumerator_next_data(e);
	      if(m != NULL)
		padico_print("Puk mod = %s; used_by = %d; manages = %d\n", m->mod_name, m->used_by, m->managed_instances);
	    }
	  padico_print("Puk: %d modules remaining\n", puk_hashtable_size(puk_mod.modl));
	  /* abort(); */
	  break;
	}
      padico_print("Puk: unloading %s\n", m->mod_name);
      padico_puk_mod_unload(m);
      padico_print("Puk: done.\n");
      puk_hashtable_enumerator_delete(e);
    }
  puk_hashtable_delete(puk_mod.modl);
  puk_lock_release();
}

